from enum import Enum
from typing import Literal, Optional

from pydantic import BaseModel, Field

from bizon.connectors.destinations.bigquery.src.config import BigQueryRecordSchemaConfig
from bizon.destination.config import (
    AbstractDestinationConfig,
    AbstractDestinationDetailsConfig,
    DestinationTypes,
)


class TimePartitioningWindow(str, Enum):
    DAY = "DAY"
    HOUR = "HOUR"
    MONTH = "MONTH"
    YEAR = "YEAR"


class TimePartitioning(BaseModel):
    type: TimePartitioningWindow = Field(default=TimePartitioningWindow.DAY, description="Time partitioning type")
    field: Optional[str] = Field(
        "_bizon_loaded_at", description="Field to partition by. You can use a transformation to create this field."
    )


class BigQueryAuthentication(BaseModel):
    service_account_key: str = Field(
        description="Service Account Key JSON string. If empty it will be infered",
        default="",
    )


class BigQueryStreamingV2ConfigDetails(AbstractDestinationDetailsConfig):
    project_id: str
    dataset_id: str
    dataset_location: Optional[str] = "US"
    time_partitioning: Optional[TimePartitioning] = Field(
        default=TimePartitioning(type=TimePartitioningWindow.DAY, field="_bizon_loaded_at"),
        description="BigQuery Time partitioning type",
    )
    authentication: Optional[BigQueryAuthentication] = None
    bq_max_rows_per_request: Optional[int] = Field(
        5000,
        description="Max rows per buffer streaming request. Must not exceed 10000.",
        le=10000,
    )
    record_schemas: Optional[list[BigQueryRecordSchemaConfig]] = Field(
        default=None, description="Schema for the records. Required if unnest is set to true."
    )


class BigQueryStreamingV2Config(AbstractDestinationConfig):
    name: Literal[DestinationTypes.BIGQUERY_STREAMING_V2]
    alias: str = "bigquery"
    config: BigQueryStreamingV2ConfigDetails



import pytest
from google.cloud.bigquery import SchemaField
from google.protobuf.json_format import ParseError
from google.protobuf.message import EncodeError

from bizon.connectors.destinations.bigquery_streaming_v2.src.destination import (
    BigQueryStreamingV2Destination,
)
from bizon.connectors.destinations.bigquery_streaming_v2.src.proto_utils import (
    get_proto_schema_and_class,
)


def test_get_proto_schema_and_class():
    bq_schema = [
        SchemaField(name="name", field_type="STRING", mode="REQUIRED"),
        SchemaField(name="age", field_type="INTEGER", mode="REQUIRED"),
    ]
    proto_schema, table_row_class = get_proto_schema_and_class(bq_schema)
    assert proto_schema is not None


def test_to_protobuf_serialization():
    # Test to_protobuf_serialization
    bq_schema = [
        SchemaField(name="name", field_type="STRING", mode="REQUIRED"),
        SchemaField(name="age", field_type="INTEGER", mode="REQUIRED"),
    ]

    proto_schema, table_row_class = get_proto_schema_and_class(bq_schema)

    data = {
        "name": "John",
        "age": 30,
    }

    serialized_record = BigQueryStreamingV2Destination.to_protobuf_serialization(table_row_class, data)

    assert serialized_record is not None


def test_to_protobuf_serialization_error_mismatch_schema():
    # Test to_protobuf_serialization
    bq_schema = [
        SchemaField(name="name", field_type="STRING", mode="REQUIRED"),
        SchemaField(name="age", field_type="INTEGER", mode="REQUIRED"),
        SchemaField(name="email", field_type="STRING", mode="REQUIRED"),
    ]

    proto_schema, table_row_class = get_proto_schema_and_class(bq_schema)

    data = {
        "name": "John",
        "age": 30,
    }

    with pytest.raises(EncodeError):
        BigQueryStreamingV2Destination.to_protobuf_serialization(table_row_class, data)


def test_to_protobuf_serialization_error_mismatch_schema_parse_error():
    # Test to_protobuf_serialization
    bq_schema = [
        SchemaField(name="name", field_type="STRING", mode="REQUIRED"),
        SchemaField(name="email", field_type="STRING", mode="REQUIRED"),
    ]

    proto_schema, table_row_class = get_proto_schema_and_class(bq_schema)

    data = {
        "not_in_schema": "John",
        "name": "John",
        "age": 30,
    }

    with pytest.raises(ParseError):
        BigQueryStreamingV2Destination.to_protobuf_serialization(table_row_class, data)



import os
import threading
from datetime import datetime
from queue import Queue

import pytest

from bizon.engine.backend.backend import AbstractBackend
from bizon.engine.backend.models import JobStatus, StreamJob
from bizon.engine.engine import RunnerFactory
from bizon.engine.pipeline.producer import Producer
from bizon.source.models import SourceIteration, SourceRecord


@pytest.fixture(scope="function")
def my_producer(my_sqlite_backend: AbstractBackend):
    runner = RunnerFactory.create_from_yaml(os.path.abspath("tests/engine/dummy_pipeline_sqlite.yml"))
    source = runner.get_source(bizon_config=runner.bizon_config, config=runner.config)
    my_sqlite_backend.create_all_tables()
    queue = runner.get_queue(bizon_config=runner.bizon_config, queue=Queue())
    return Producer(bizon_config=runner.bizon_config, queue=queue, source=source, backend=my_sqlite_backend)


@pytest.fixture(scope="function")
def my_job(my_producer: Producer, sqlite_db_session):
    job = my_producer.backend.create_stream_job(
        name=my_producer.bizon_config.name,
        source_name=my_producer.source.config.name,
        stream_name=my_producer.source.config.stream,
        sync_mode="full_refresh",
        job_status=JobStatus.STARTED,
        session=sqlite_db_session,
    )
    return job


def test_cursor_recovery_after_iteration(my_producer: Producer, sqlite_db_session, my_job: StreamJob):
    cursor = my_producer.get_or_create_cursor(job_id=my_job.id, session=sqlite_db_session)
    assert cursor is not None

    stop_event = threading.Event()
    my_producer.run(job_id=my_job.id, stop_event=stop_event)

    # Here we did not run the job, so the cursor should be None
    cursor_from_db = my_producer.backend.get_last_cursor_by_job_id(job_id=my_job.id, session=sqlite_db_session)
    assert cursor_from_db is None


def test_cursor_recovery_not_running(my_producer: Producer, sqlite_db_session, my_job: StreamJob):
    cursor = my_producer.get_or_create_cursor(job_id=my_job.id, session=sqlite_db_session)
    assert cursor is not None

    # Here we did not run the job, so the cursor should be None
    cursor_from_db = my_producer.backend.get_last_cursor_by_job_id(job_id=my_job.id, session=sqlite_db_session)
    assert cursor_from_db is None


def test_source_thread(my_producer: Producer):
    assert my_producer.source.config.name == "dummy"


def test_queue_is_full(my_producer: Producer, sqlite_db_session, my_job: StreamJob):
    assert my_producer.queue.config.max_nb_messages == 1_000_000

    my_producer.queue.connect()

    cursor = my_producer.get_or_create_cursor(job_id=my_job.id, session=sqlite_db_session)

    is_queue_full, queue_size, approximate_nb_records_in_queue = my_producer.is_queue_full(cursor)
    assert is_queue_full is False
    assert queue_size == 0
    assert approximate_nb_records_in_queue == 0

    source_iteration = SourceIteration(
        records=[
            SourceRecord(
                id=str(i),
                data={"test": "test"},
                timestamp="2021-01-01T00:00:00",
            )
            for i in range(1000)
        ],
        next_pagination={"test": "test"},
    )

    my_producer.queue.put(
        source_iteration=source_iteration,
        iteration=cursor.iteration,
        extracted_at=datetime.now(),
    )

    cursor.update_state(
        pagination_dict=source_iteration.next_pagination, nb_records_fetched=len(source_iteration.records)
    )

    is_queue_full, queue_size, approximate_nb_records_in_queue = my_producer.is_queue_full(cursor)
    assert is_queue_full is False
    assert queue_size == 1
    assert approximate_nb_records_in_queue == 1000

    for i in range(1000):
        my_producer.queue.put(
            source_iteration=source_iteration,
            iteration=cursor.iteration,
            extracted_at=datetime.now(),
        )
        cursor.update_state(
            pagination_dict=source_iteration.next_pagination, nb_records_fetched=len(source_iteration.records)
        )

    is_queue_full, queue_size, approximate_nb_records_in_queue = my_producer.is_queue_full(cursor)
    assert is_queue_full is True
    assert queue_size == 1001
    assert approximate_nb_records_in_queue == 1_001_000



import pytest

from bizon.engine.backend.models import JobStatus
from bizon.source.cursor import Cursor


@pytest.fixture
def my_cursor():
    return Cursor(
        source_name="dummy",
        stream_name="test",
        job_id="jfd87djkj",
        total_records=100,
    )


def test_cursor_instance(my_cursor: Cursor):
    assert my_cursor.source_name == "dummy"
    assert my_cursor.stream_name == "test"
    assert my_cursor.source_full_name == "dummy.test"
    assert my_cursor.total_records == 100
    assert my_cursor.iteration == 0
    assert my_cursor.job_status == JobStatus.STARTED
    assert my_cursor.is_finished == False


def test_cursor_percentage(my_cursor: Cursor):
    my_cursor.update_state(
        pagination_dict={"after": "abc"},
        nb_records_fetched=20,
    )
    assert my_cursor.percentage_fetched == 0.2


def test_update_state_pagination_empty_after_first_iteration(my_cursor: Cursor):
    my_cursor.update_state(
        pagination_dict={},
        nb_records_fetched=20,
    )
    assert my_cursor.job_status == JobStatus.SUCCEEDED
    assert my_cursor.is_finished == True


def test_update_state_pagination_empty_after_two_iterations(my_cursor: Cursor):
    my_cursor.update_state(
        pagination_dict={"after": "abc"},
        nb_records_fetched=20,
    )

    assert my_cursor.job_status == JobStatus.RUNNING
    assert my_cursor.iteration == 1
    assert my_cursor.is_finished == False

    my_cursor.update_state(
        pagination_dict={},
        nb_records_fetched=20,
    )

    assert my_cursor.job_status == JobStatus.SUCCEEDED
    assert my_cursor.is_finished == True
    assert my_cursor.iteration == 2



import os
import tempfile

from click.testing import CliRunner

from bizon.cli.main import cli

test_pg_config = f"""
name: test_pipeline

source:
  name: dummy
  stream: creatures
  authentication:
    type: api_key
    params:
      token: dummy_key

destination:
  name: logger
  config:
    dummy: dummy

engine:
  backend:
    type: postgres
    config:
        database: bizon_test
        schema: public
        syncCursorInDBEvery: 2
        host: {os.environ.get("POSTGRES_HOST", "localhost")}
        port: 5432
        username: postgres
        password: bizon
"""


def test_run_command():
    with tempfile.NamedTemporaryFile(delete=False) as temp:
        # Write config in temp file
        with open(temp.name, "w") as f:
            f.write(test_pg_config)

        runner = CliRunner()
        result = runner.invoke(cli, ["run", temp.name])
        assert result.exit_code == 0


def test_run_command_debug():
    with tempfile.NamedTemporaryFile(delete=False) as temp:
        # Write config in temp file
        with open(temp.name, "w") as f:
            f.write(test_pg_config)

        runner = CliRunner()
        result = runner.invoke(cli, ["run", temp.name, "--log-level", "DEBUG"])
        assert result.exit_code == 0


def test_source_list_command():
    runner = CliRunner()
    result = runner.invoke(cli, ["source", "list"])
    assert "Available sources:" in result.output
    assert "dummy" in result.output
    assert result.exit_code == 0


def test_stream_list_command():
    runner = CliRunner()
    result = runner.invoke(cli, ["stream", "list", "dummy"])
    assert "Available streams for dummy:" in result.output
    assert "[Full refresh only] - plants" in result.output
    assert result.exit_code == 0



import pytest
from pydantic import ValidationError

from bizon.destination.config import (
    AbstractDestinationConfig,
    AbstractDestinationDetailsConfig,
    RecordSchemaConfig,
)


def test_config():
    config = AbstractDestinationConfig(
        name="file",
        alias="file",
        config=AbstractDestinationDetailsConfig(
            unnest=False,
        ),
    )
    assert config


def test_config_no_record_schema_provided():
    with pytest.raises(ValidationError) as e:
        AbstractDestinationConfig(
            name="file",
            alias="file",
            config=AbstractDestinationDetailsConfig(
                unnest=True,
            ),
        )


def test_config_with_unnest_provided_schema():
    config = AbstractDestinationConfig(
        name="file",
        alias="file",
        config=AbstractDestinationDetailsConfig(
            unnest=True,
            record_schemas=[
                RecordSchemaConfig(
                    destination_id="cookie",
                    record_schema=[
                        {"name": "name", "type": "string", "description": "Name of the user"},
                        {"name": "age", "type": "int", "description": "Age of the user"},
                    ],
                ),
            ],
        ),
    )
    assert config



import os
import tempfile
from datetime import datetime
from typing import List, Tuple

import orjson
import polars as pl
import urllib3.exceptions
from google.api_core.exceptions import (
    Conflict,
    NotFound,
    RetryError,
    ServerError,
    ServiceUnavailable,
)
from google.cloud import bigquery, bigquery_storage_v1
from google.cloud.bigquery import DatasetReference, TimePartitioning
from google.cloud.bigquery_storage_v1.types import (
    AppendRowsRequest,
    ProtoRows,
    ProtoSchema,
)
from loguru import logger
from requests.exceptions import ConnectionError, SSLError, Timeout
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from bizon.common.models import SyncMetadata
from bizon.connectors.destinations.bigquery.src.config import (
    BigQueryColumnMode,
    BigQueryColumnType,
)
from bizon.destination.destination import AbstractDestination
from bizon.engine.backend.backend import AbstractBackend
from bizon.monitoring.monitor import AbstractMonitor
from bizon.source.callback import AbstractSourceCallback

from .config import BigQueryStreamingConfigDetails


class BigQueryStreamingDestination(AbstractDestination):
    # Add constants for limits
    MAX_REQUEST_SIZE_BYTES = 5 * 1024 * 1024  # 5 MB (max is 10MB)
    MAX_ROW_SIZE_BYTES = 0.9 * 1024 * 1024  # 1 MB

    def __init__(
        self,
        sync_metadata: SyncMetadata,
        config: BigQueryStreamingConfigDetails,
        backend: AbstractBackend,
        source_callback: AbstractSourceCallback,
        monitor: AbstractMonitor,
    ):  # type: ignore
        super().__init__(sync_metadata, config, backend, source_callback, monitor)
        self.config: BigQueryStreamingConfigDetails = config

        if config.authentication and config.authentication.service_account_key:
            with tempfile.NamedTemporaryFile(delete=False) as temp:
                temp.write(config.authentication.service_account_key.encode())
                temp_file_path = temp.name
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = temp_file_path

        self.project_id = config.project_id
        self.bq_client = bigquery.Client(project=self.project_id)
        self.bq_storage_client = bigquery_storage_v1.BigQueryWriteClient()
        self.dataset_id = config.dataset_id
        self.dataset_location = config.dataset_location
        self.bq_max_rows_per_request = config.bq_max_rows_per_request

    @property
    def table_id(self) -> str:
        tabled_id = f"{self.sync_metadata.source_name}_{self.sync_metadata.stream_name}"
        return self.destination_id or f"{self.project_id}.{self.dataset_id}.{tabled_id}"

    def get_bigquery_schema(self) -> List[bigquery.SchemaField]:
        if self.config.unnest:
            if len(list(self.record_schemas.keys())) == 1:
                self.destination_id = list(self.record_schemas.keys())[0]

            return [
                bigquery.SchemaField(
                    name=col.name,
                    field_type=col.type,
                    mode=col.mode,
                    description=col.description,
                    default_value_expression=col.default_value_expression,
                )
                for col in self.record_schemas[self.destination_id]
            ]

        # Case we don't unnest the data
        else:
            return [
                bigquery.SchemaField(
                    "_source_record_id",
                    BigQueryColumnType.STRING,
                    mode=BigQueryColumnMode.REQUIRED,
                    description="The source record id",
                ),
                bigquery.SchemaField(
                    "_source_timestamp",
                    BigQueryColumnType.TIMESTAMP,
                    mode=BigQueryColumnMode.REQUIRED,
                    description="The source timestamp",
                ),
                bigquery.SchemaField(
                    "_source_data",
                    BigQueryColumnType.JSON,
                    mode=BigQueryColumnMode.NULLABLE,
                    description="The source data",
                ),
                bigquery.SchemaField(
                    "_bizon_extracted_at",
                    BigQueryColumnType.TIMESTAMP,
                    mode=BigQueryColumnMode.REQUIRED,
                    description="The bizon extracted at",
                ),
                bigquery.SchemaField(
                    "_bizon_loaded_at",
                    BigQueryColumnType.TIMESTAMP,
                    mode=BigQueryColumnMode.REQUIRED,
                    default_value_expression="CURRENT_TIMESTAMP()",
                    description="The bizon loaded at",
                ),
                bigquery.SchemaField(
                    "_bizon_id",
                    BigQueryColumnType.STRING,
                    mode=BigQueryColumnMode.REQUIRED,
                    description="The bizon id",
                ),
            ]

    def check_connection(self) -> bool:
        dataset_ref = DatasetReference(self.project_id, self.dataset_id)

        try:
            self.bq_client.get_dataset(dataset_ref)
        except NotFound:
            dataset = bigquery.Dataset(dataset_ref)
            dataset.location = self.dataset_location
            dataset = self.bq_client.create_dataset(dataset)
        return True

    def append_rows_to_stream(
        self,
        write_client: bigquery_storage_v1.BigQueryWriteClient,
        stream_name: str,
        proto_schema: ProtoSchema,
        serialized_rows: List[bytes],
    ):
        request = AppendRowsRequest(
            write_stream=stream_name,
            proto_rows=AppendRowsRequest.ProtoData(
                rows=ProtoRows(serialized_rows=serialized_rows),
                writer_schema=proto_schema,
            ),
        )
        response = write_client.append_rows(iter([request]))
        return response.code().name

    def safe_cast_record_values(self, row: dict):
        """
        Safe cast record values to the correct type for BigQuery.
        """
        for col in self.record_schemas[self.destination_id]:
            # Handle dicts as strings
            if col.type in [BigQueryColumnType.STRING, BigQueryColumnType.JSON]:
                if isinstance(row[col.name], dict) or isinstance(row[col.name], list):
                    row[col.name] = orjson.dumps(row[col.name]).decode("utf-8")

            # Handle timestamps
            if (
                col.type in [BigQueryColumnType.TIMESTAMP, BigQueryColumnType.DATETIME]
                and col.default_value_expression is None
            ):
                if isinstance(row[col.name], int):
                    if row[col.name] > datetime(9999, 12, 31).timestamp():
                        row[col.name] = datetime.fromtimestamp(row[col.name] / 1_000_000).strftime(
                            "%Y-%m-%d %H:%M:%S.%f"
                        )
                    else:
                        try:
                            row[col.name] = datetime.fromtimestamp(row[col.name]).strftime("%Y-%m-%d %H:%M:%S.%f")
                        except ValueError:
                            error_message = (
                                f"Error casting timestamp for destination '{self.destination_id}' column '{col.name}'. "
                                f"Invalid timestamp value: {row[col.name]} ({type(row[col.name])}). "
                                "Consider using a transformation."
                            )
                            logger.error(error_message)
                            raise ValueError(error_message)
        return row

    @retry(
        retry=retry_if_exception_type(
            (
                ServerError,
                ServiceUnavailable,
                SSLError,
                ConnectionError,
                Timeout,
                RetryError,
                urllib3.exceptions.ProtocolError,
                urllib3.exceptions.SSLError,
            )
        ),
        wait=wait_exponential(multiplier=2, min=4, max=120),
        stop=stop_after_attempt(8),
        before_sleep=lambda retry_state: logger.warning(
            f"Attempt {retry_state.attempt_number} failed. Retrying in {retry_state.next_action.sleep} seconds..."
        ),
    )
    def _insert_batch(self, table, batch):
        """Helper method to insert a batch of rows with retry logic"""
        logger.debug(f"Inserting batch in table {table.table_id}")
        try:
            # Handle streaming batch
            if batch.get("stream_batch") and len(batch["stream_batch"]) > 0:
                self.bq_client.insert_rows_json(
                    table,
                    batch["stream_batch"],
                    row_ids=[None] * len(batch["stream_batch"]),
                    timeout=300,  # 5 minutes timeout per request
                )

            # Handle large rows batch
            if batch.get("json_batch") and len(batch["json_batch"]) > 0:
                job_config = bigquery.LoadJobConfig(
                    source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
                    schema=table.schema,
                    ignore_unknown_values=True,
                )

                load_job = self.bq_client.load_table_from_json(
                    batch["json_batch"], table, job_config=job_config, timeout=300
                )
                load_job.result()

                if load_job.state != "DONE":
                    raise Exception(f"Failed to load rows to BigQuery: {load_job.errors}")

                self.monitor.track_large_records_synced(
                    num_records=len(batch["json_batch"]), extra_tags={"destination_id": self.destination_id}
                )

        except Exception as e:
            logger.error(f"Error inserting batch: {str(e)}, type: {type(e)}")
            raise

    def load_to_bigquery_via_legacy_streaming(self, df_destination_records: pl.DataFrame) -> str:
        # Create table if it does not exist
        schema = self.get_bigquery_schema()
        table = bigquery.Table(self.table_id, schema=schema)
        time_partitioning = TimePartitioning(
            field=self.config.time_partitioning.field, type_=self.config.time_partitioning.type
        )
        table.time_partitioning = time_partitioning

        if self.clustering_keys and self.clustering_keys[self.destination_id]:
            table.clustering_fields = self.clustering_keys[self.destination_id]
        try:
            table = self.bq_client.create_table(table)
        except Conflict:
            table = self.bq_client.get_table(self.table_id)
            # Compare and update schema if needed
            existing_fields = {field.name: field for field in table.schema}
            new_fields = {field.name: field for field in self.get_bigquery_schema()}

            # Find fields that need to be added
            fields_to_add = [field for name, field in new_fields.items() if name not in existing_fields]

            if fields_to_add:
                logger.warning(f"Adding new fields to table schema: {[field.name for field in fields_to_add]}")
                updated_schema = table.schema + fields_to_add
                table.schema = updated_schema
                table = self.bq_client.update_table(table, ["schema"])

        if self.config.unnest:
            # We cannot use the `json_decode` method here because of the issue: https://github.com/pola-rs/polars/issues/22371
            rows_to_insert = [
                self.safe_cast_record_values(orjson.loads(row))
                for row in df_destination_records["source_data"].to_list()
            ]
        else:
            df_destination_records = df_destination_records.with_columns(
                pl.col("bizon_extracted_at").dt.strftime("%Y-%m-%d %H:%M:%S").alias("bizon_extracted_at"),
                pl.col("bizon_loaded_at").dt.strftime("%Y-%m-%d %H:%M:%S").alias("bizon_loaded_at"),
                pl.col("source_timestamp").dt.strftime("%Y-%m-%d %H:%M:%S").alias("source_timestamp"),
            )
            df_destination_records = df_destination_records.rename(
                {
                    "bizon_id": "_bizon_id",
                    "bizon_extracted_at": "_bizon_extracted_at",
                    "bizon_loaded_at": "_bizon_loaded_at",
                    "source_record_id": "_source_record_id",
                    "source_timestamp": "_source_timestamp",
                    "source_data": "_source_data",
                }
            )
            rows_to_insert = [row for row in df_destination_records.iter_rows(named=True)]

        errors = []
        for batch in self.batch(rows_to_insert):
            try:
                batch_errors = self._insert_batch(table, batch)
                if batch_errors:
                    errors.extend(batch_errors)
            except Exception as e:
                logger.error(f"Failed to insert batch on destination {self.destination_id} after all retries: {str(e)}")
                if isinstance(e, RetryError):
                    logger.error(f"Retry error details: {e.cause if hasattr(e, 'cause') else 'No cause available'}")
                raise

        if errors:
            logger.error("Encountered errors while inserting rows:")
            for error in errors:
                if error.get("errors") and len(error["errors"]) > 0:
                    logger.error("The following row failed to be inserted:")
                    if batch.get("stream_batch") and len(batch["stream_batch"]) > 0:
                        logger.error(f"{batch['stream_batch'][error['index']]}")
                    else:
                        logger.error(f"{batch['json_batch'][error['index']]}")
                    for error_detail in error["errors"]:
                        logger.error(f"Location (column): {error_detail['location']}")
                        logger.error(f"Reason: {error_detail['reason']}")
                        logger.error(f"Message: {error_detail['message']}")
            raise Exception(f"Encountered errors while inserting rows: {errors}")

    def write_records(self, df_destination_records: pl.DataFrame) -> Tuple[bool, str]:
        logger.debug("Using BigQuery legacy streaming API...")
        self.load_to_bigquery_via_legacy_streaming(df_destination_records=df_destination_records)
        return True, ""

    def batch(self, iterable):
        """
        Yield successive batches respecting both row count and size limits.
        """
        current_batch = []
        current_batch_size = 0
        large_rows = []

        for item in iterable:
            # Estimate the size of the item (as JSON)
            item_size = len(str(item).encode("utf-8"))

            # If adding this item would exceed either limit, yield current batch and start new one
            if (
                len(current_batch) >= self.bq_max_rows_per_request
                or current_batch_size + item_size > self.MAX_REQUEST_SIZE_BYTES
            ):
                logger.debug(
                    f"Yielding batch of {len(current_batch)} rows, size: {current_batch_size / 1024 / 1024:.2f}MB"
                )
                yield {"stream_batch": current_batch, "json_batch": large_rows}
                current_batch = []
                current_batch_size = 0
                large_rows = []

            if item_size > self.MAX_ROW_SIZE_BYTES:
                large_rows.append(item)
                logger.debug(f"Large row detected: {item_size} bytes")
            else:
                current_batch.append(item)
                current_batch_size += item_size

        # Yield the last batch
        if current_batch:
            logger.debug(
                f"Yielding streaming batch of {len(current_batch)} rows, size: {current_batch_size / 1024 / 1024:.2f}MB"
            )
            logger.debug(f"Yielding large rows batch of {len(large_rows)} rows")
            yield {"stream_batch": current_batch, "json_batch": large_rows}



import pytest

from bizon.connectors.sources.dummy.src.source import (
    DummyAuthConfig,
    DummySource,
    DummySourceConfig,
)
from bizon.source.auth.authenticators.token import TokenAuthParams
from bizon.source.auth.config import AuthType


@pytest.fixture
def dummy_source_config():
    return DummySourceConfig(
        name="dummy",
        stream="creatures",
        authentication=DummyAuthConfig(type=AuthType.API_KEY, params=TokenAuthParams(token="dummy_key")),
    )


@pytest.fixture
def dummy_source(dummy_source_config):
    return DummySource(config=dummy_source_config)


def test_dummy_source_url_entity(dummy_source: DummySource):
    assert dummy_source.url_entity == "https://api.dummy.com/v1/creatures"


def test_dummy_source_get_auth(dummy_source: DummySource):
    assert dummy_source.get_authenticator().token == "Bearer dummy_key"


def test_get_first_page(dummy_source: DummySource):
    source_iteration = dummy_source.get()
    assert len(source_iteration.records) == 2
    assert source_iteration.next_pagination == {"cursor": "vfvfvuhfefpeiduzhihxb"}


def test_get_last_page(dummy_source: DummySource):
    source_iteration = dummy_source.get(pagination={"cursor": "final-cursor"})
    assert len(source_iteration.records) == 1
    assert source_iteration.next_pagination == {}



import os

import pytest

from bizon.common.models import SyncMetadata
from bizon.connectors.destinations.bigquery.src.config import (
    BigQueryConfig,
    BigQueryConfigDetails,
    GCSBufferFormat,
)
from bizon.connectors.destinations.bigquery.src.destination import BigQueryDestination
from bizon.destination.config import DestinationTypes
from bizon.destination.destination import DestinationFactory
from bizon.monitoring.noop.monitor import NoOpMonitor


@pytest.fixture(scope="function")
def sync_metadata() -> SyncMetadata:
    return SyncMetadata(
        job_id="rfou98C9DJH",
        source_name="cookie",
        stream_name="test",
        destination_name="bigquery",
        destination_alias="bigquery",
        sync_mode="full_refresh",
    )


@pytest.mark.skipif(
    os.getenv("POETRY_ENV_TEST") == "CI",
    reason="Skipping tests that require a BigQuery database",
)
def test_bigquery_factory(sync_metadata, my_backend):
    config = BigQueryConfig(
        name=DestinationTypes.BIGQUERY,
        config=BigQueryConfigDetails(
            project_id="project_id",
            dataset_id="dataset_id",
            table_id="table_id",
            credentials_path="credentials_path",
            gcs_buffer_bucket="gcs_buffer_bucket",
            gcs_buffer_format=GCSBufferFormat.PARQUET,
            authentication={"service_account_key": ""},
        ),
    )

    destination = DestinationFactory().get_destination(
        sync_metadata=sync_metadata,
        config=config,
        backend=my_backend,
        monitor=NoOpMonitor(sync_metadata=sync_metadata, monitoring_config=None),
    )
    assert isinstance(destination, BigQueryDestination)
    assert destination.config.authentication.service_account_key == ""
    assert destination.config.project_id == "project_id"
    assert destination.config.dataset_id == "dataset_id"


@pytest.mark.skipif(
    os.getenv("POETRY_ENV_TEST") == "CI",
    reason="Skipping tests that require a BigQuery database",
)
def test_bigquery_factory_empty_service_account(sync_metadata, my_backend):
    config = BigQueryConfig(
        name=DestinationTypes.BIGQUERY,
        config=BigQueryConfigDetails(
            project_id="project_id",
            dataset_id="dataset_id",
            table_id="table_id",
            credentials_path="credentials_path",
            gcs_buffer_bucket="gcs_buffer_bucket",
            gcs_buffer_format=GCSBufferFormat.PARQUET,
        ),
    )

    destination = DestinationFactory().get_destination(
        sync_metadata=sync_metadata,
        config=config,
        backend=my_backend,
        monitor=NoOpMonitor(sync_metadata=sync_metadata, monitoring_config=None),
    )
    assert isinstance(destination, BigQueryDestination)
    assert destination.config.authentication is None



import os
from queue import Queue

import pytest

from bizon.engine.backend.adapters.sqlalchemy.backend import SQLAlchemyBackend
from bizon.engine.backend.backend import AbstractBackend
from bizon.engine.backend.models import JobStatus
from bizon.engine.engine import RunnerFactory, replace_env_variables_in_config
from bizon.engine.pipeline.producer import Producer
from bizon.engine.runner.adapters.thread import ThreadRunner


def test_load_from_config():
    runner = RunnerFactory.create_from_yaml(os.path.abspath("tests/engine/dummy_pipeline_sqlite.yml"))
    assert runner.is_running is False


def test_replace_env_variables_in_config():
    os.environ["BIZON_ENV_SOURCE_URL"] = "https://dummy.com"
    config = {
        "source": {
            "name": "dummy",
            "url": "BIZON_ENV_SOURCE_URL",
        }
    }
    config = replace_env_variables_in_config(config=config)
    assert config["source"]["url"] == "https://dummy.com"


def test_replace_env_variables_in_yaml_config():
    os.environ["BIZON_ENV_POSTGRES_HOST"] = "localhost"
    os.environ["BIZON_ENV_POSTGRES_USERNAME"] = "postgres"
    os.environ["BIZON_ENV_POSTGRES_PASSWORD"] = "bizon"
    os.environ["BIZON_ENV_MAX_ITERATIONS"] = "10"

    runner = RunnerFactory.create_from_yaml(os.path.abspath("tests/engine/dummy_pipeline_postgres_env_variables.yml"))

    assert runner.bizon_config.engine.backend.config.host == "localhost"
    assert runner.bizon_config.engine.backend.config.username == "postgres"
    assert runner.bizon_config.engine.backend.config.password == "bizon"
    assert runner.bizon_config.source.max_iterations == 10


@pytest.fixture(scope="function")
def my_producer(my_sqlite_backend: AbstractBackend):
    runner = RunnerFactory.create_from_yaml(os.path.abspath("tests/engine/dummy_pipeline_sqlite.yml"))
    source = runner.get_source(bizon_config=runner.bizon_config, config=runner.config)
    my_sqlite_backend.create_all_tables()
    source_backend = my_sqlite_backend
    queue = runner.get_queue(bizon_config=runner.bizon_config, queue=Queue())
    return Producer(bizon_config=runner.bizon_config, queue=queue, source=source, backend=source_backend)


@pytest.fixture(scope="function")
def my_runner(my_sqlite_backend) -> ThreadRunner:
    runner = RunnerFactory.create_from_yaml(os.path.abspath("tests/engine/dummy_pipeline_sqlite.yml"))
    runner.backend = my_sqlite_backend
    runner.backend.create_all_tables()
    return runner


def test_run(my_runner: ThreadRunner):
    assert my_runner.is_running is False


def test_get_source(my_runner: ThreadRunner):
    source = my_runner.get_source(bizon_config=my_runner.bizon_config, config=my_runner.config)
    assert source.config.name == "dummy"


def test_get_backend(my_runner: ThreadRunner):
    backend: SQLAlchemyBackend = my_runner.get_backend(bizon_config=my_runner.bizon_config)
    assert backend.type == "sqlite_in_memory"


def test_create_job(my_runner: ThreadRunner, sqlite_db_session):
    job = my_runner.get_or_create_job(
        bizon_config=my_runner.bizon_config,
        backend=my_runner.backend,
        source=my_runner.get_source(bizon_config=my_runner.bizon_config, config=my_runner.config),
        session=sqlite_db_session,
    )
    assert job is not None

    my_job = my_runner.backend.get_stream_job_by_id(job_id=job.id, session=sqlite_db_session)
    assert my_job.id == job.id


def test_create_job_and_recover(my_runner: ThreadRunner, sqlite_db_session):
    source = my_runner.get_source(bizon_config=my_runner.bizon_config, config=my_runner.config)

    job = my_runner.get_or_create_job(
        bizon_config=my_runner.bizon_config, backend=my_runner.backend, source=source, session=sqlite_db_session
    )
    assert job is not None

    my_runner.backend.update_stream_job_status(job_id=job.id, job_status=JobStatus.RUNNING, session=sqlite_db_session)

    recover_job = my_runner.get_or_create_job(
        bizon_config=my_runner.bizon_config, backend=my_runner.backend, source=source, session=sqlite_db_session
    )
    assert recover_job.id == job.id


def test_source_thread(my_producer: Producer):
    assert my_producer.source.config.name == "dummy"



import multiprocessing
import multiprocessing.synchronize
import threading
from typing import Union

from loguru import logger

from bizon.destination.destination import AbstractDestination
from bizon.engine.pipeline.consumer import AbstractQueueConsumer
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.engine.queue.config import QueueMessage
from bizon.engine.queue.queue import AbstractQueue
from bizon.monitoring.monitor import AbstractMonitor
from bizon.transform.transform import Transform

from .config import PythonQueueConfig


class PythonQueueConsumer(AbstractQueueConsumer):
    def __init__(
        self,
        config: PythonQueueConfig,
        queue: AbstractQueue,
        destination: AbstractDestination,
        transform: Transform,
        monitor: AbstractMonitor,
    ):
        super().__init__(
            config,
            destination=destination,
            transform=transform,
            monitor=monitor,
        )
        self.queue = queue
        self.monitor.track_pipeline_status(PipelineReturnStatus.RUNNING)

    def run(self, stop_event: Union[threading.Event, multiprocessing.synchronize.Event]) -> PipelineReturnStatus:
        while True:
            # Handle kill signal from the runner
            if stop_event.is_set():
                logger.info("Stop event is set, closing consumer ...")
                self.monitor.track_pipeline_status(PipelineReturnStatus.KILLED_BY_RUNNER)
                return PipelineReturnStatus.KILLED_BY_RUNNER

            # Retrieve the message from the queue
            queue_message: QueueMessage = self.queue.get()

            status = self.process_queue_message(queue_message)

            if status != PipelineReturnStatus.RUNNING:
                self.queue.task_done()
                return status



import base64

from pydantic import BaseModel, Field

from .abstract_token import AbstractHeaderAuthenticator


class BasicHttpAuthParams(BaseModel):
    username: str = Field(..., description="Username for basic auth")
    password: str = Field("", description="Password for basic auth")
    auth_method: str = Field("Basic", description="Auth method for basic auth")
    auth_header: str = Field("Authorization", description="Auth header for basic auth")


class BasicHttpAuthenticator(AbstractHeaderAuthenticator):
    """
    Builds auth based off the basic authentication scheme as defined by RFC 7617, which transmits credentials as USER ID/password pairs, encoded using bas64
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Authentication#basic_authentication_scheme
    """

    @property
    def auth_header(self) -> str:
        return self._auth_header

    @property
    def token(self) -> str:
        return f"{self._auth_method} {self._token}"

    def __init__(self, params: BasicHttpAuthParams):
        auth_string = f"{params.username}:{params.password}".encode()
        b64_encoded = base64.b64encode(auth_string).decode("utf8")
        self._auth_header = params.auth_header
        self._auth_method = params.auth_method
        self._token = b64_encoded



import uuid

import pytest
from sqlalchemy.orm import Session

from bizon.engine.backend.adapters.sqlalchemy.backend import SQLAlchemyBackend
from bizon.engine.backend.models import CursorStatus, JobStatus, SourceCursor


@pytest.mark.parametrize("backend", [pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("my_sqlite_backend")])
def test_backend_client_instance(backend: SQLAlchemyBackend):
    assert backend.config.syncCursorInDBEvery == 2


@pytest.mark.parametrize("backend", [pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("my_sqlite_backend")])
def test_create_all_then_check_prerequisites(backend: SQLAlchemyBackend):
    backend.create_all_tables()
    all_entities_exist = backend.check_prerequisites()
    assert all_entities_exist == True


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_create_stream_job(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()

    new_job_id = backend.create_stream_job(
        session=session,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.STARTED,
    )

    assert new_job_id


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_update_stream_job_status(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()

    job_name = f"job_{uuid.uuid4().hex}"

    new_job = backend.create_stream_job(
        session=session,
        name=job_name,
        source_name="sourcetest",
        stream_name="cookie",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.STARTED,
    )

    # If no job is running, get_running_job_id should return None
    job_running_not_found = backend.get_running_stream_job(
        session=session, name=job_name, source_name="sourcetest", stream_name="cookie"
    )
    assert job_running_not_found == None

    backend.update_stream_job_status(session=session, job_id=new_job.id, job_status=JobStatus.RUNNING)

    job_running = backend.get_running_stream_job(
        session=session, name=job_name, source_name="sourcetest", stream_name="cookie"
    )
    assert new_job.id == job_running.id


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_stream_job_create_and_get(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()

    new_job = backend.create_stream_job(
        session=session,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.STARTED,
    )

    job = backend.get_stream_job_by_id(session=session, job_id=new_job.id)
    assert job.id == new_job.id


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_create_source_cursor(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()

    new_job = backend.create_stream_job(
        session=session,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.STARTED,
    )

    new_cursor_id = backend.create_source_cursor(
        session=session,
        job_id=new_job.id,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        iteration=0,
        rows_fetched=0,
        next_pagination=None,
        cursor_status=CursorStatus.STARTED,
    )

    assert new_cursor_id


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_create_failed_cursor(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()

    new_job = backend.create_stream_job(
        session=session,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.RUNNING,
    )

    new_cursor_id = backend.create_source_cursor(
        session=session,
        job_id=new_job.id,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        iteration=0,
        rows_fetched=0,
        next_pagination=None,
        cursor_status=CursorStatus.FAILED,
        error_message="This is a test error message",
    )

    assert new_cursor_id


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_update_source_cursor(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()

    new_job = backend.create_stream_job(
        session=session,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.RUNNING,
    )

    new_cursor_id = backend.create_source_cursor(
        session=session,
        job_id=new_job.id,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        iteration=0,
        rows_fetched=0,
        next_pagination=None,
        cursor_status=CursorStatus.STARTED,
    )

    backend.update_source_cursor_status(
        session=session,
        cursor_id=new_cursor_id,
        cursor_status=CursorStatus.WRITTEN_IN_DESTINATION,
    )

    cursor = session.query(SourceCursor).filter_by(id=new_cursor_id).first()
    assert cursor.status == CursorStatus.WRITTEN_IN_DESTINATION


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_get_last_iteration_cursor(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()

    new_job = backend.create_stream_job(
        session=session,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.RUNNING,
    )

    cursor = backend.get_last_cursor_by_job_id(job_id=new_job.id, session=session)
    assert cursor is None

    backend.create_destination_cursor(
        session=session,
        job_id=new_job.id,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        destination_name="destinationtest",
        from_source_iteration=0,
        to_source_iteration=0,
        rows_written=0,
        success=True,
    )

    cursor = backend.get_last_cursor_by_job_id(job_id=new_job.id, session=session)

    assert cursor.job_id == new_job.id


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_create_destination_cursor(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()

    new_job = backend.create_stream_job(
        session=session,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.RUNNING,
    )

    new_destination_cursor = backend.create_destination_cursor(
        session=session,
        job_id=new_job.id,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        destination_name="destinationtest",
        from_source_iteration=0,
        to_source_iteration=0,
        rows_written=0,
        success=True,
    )

    assert new_destination_cursor.rows_written == 0


@pytest.mark.parametrize(
    "backend,session",
    [
        (pytest.lazy_fixture("my_pg_backend"), pytest.lazy_fixture("pg_db_session")),
        (pytest.lazy_fixture("my_sqlite_backend"), pytest.lazy_fixture("sqlite_db_session")),
    ],
)
def test_number_of_rows_written(backend: SQLAlchemyBackend, session: Session):
    backend.create_all_tables()
    new_job = backend.create_stream_job(
        session=session,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        sync_mode="full_refresh",
        total_records_to_fetch=200,
        job_status=JobStatus.RUNNING,
    )

    backend.create_destination_cursor(
        session=session,
        job_id=new_job.id,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        destination_name="destinationtest",
        from_source_iteration=0,
        to_source_iteration=0,
        rows_written=5,
        success=True,
    )
    backend.create_destination_cursor(
        session=session,
        job_id=new_job.id,
        name="testjob",
        source_name="sourcetest",
        stream_name="streamtest",
        destination_name="destinationtest",
        from_source_iteration=0,
        to_source_iteration=0,
        rows_written=10,
        success=True,
    )
    assert backend.get_number_of_written_rows_for_job(job_id=new_job.id, session=session) == 15



import os
import time
from datetime import datetime
from typing import List

import polars as pl
import simplejson as json
from loguru import logger
from pytz import UTC

from bizon.common.models import BizonConfig, SyncMetadata
from bizon.connectors.destinations.bigquery.src.config import BigQueryRecordSchemaConfig
from bizon.destination.models import transform_to_df_destination_records
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.engine.runner.config import RunnerStatus
from bizon.engine.runner.runner import AbstractRunner
from bizon.source.models import SourceRecord, source_record_schema
from bizon.source.source import AbstractSource


class StreamingRunner(AbstractRunner):
    def __init__(self, config: BizonConfig):
        super().__init__(config)

    @staticmethod
    def convert_source_records(records: List[SourceRecord]) -> pl.DataFrame:
        return pl.DataFrame(
            {
                "id": [record.id for record in records],
                "data": [json.dumps(record.data, ensure_ascii=False) for record in records],
                "timestamp": [record.timestamp for record in records],
                "destination_id": [record.destination_id for record in records],
            },
            schema=source_record_schema,
        )

    @staticmethod
    def convert_to_destination_records(df_source_records: pl.DataFrame, extracted_at: datetime) -> pl.DataFrame:
        return transform_to_df_destination_records(df_source_records=df_source_records, extracted_at=extracted_at)

    def _apply_streams_config(self, source: AbstractSource = None) -> None:
        """Apply streams configuration to source and destination.

        This method is completely source-agnostic. Each source connector is responsible
        for handling streams config appropriately via set_streams_config().

        When a top-level 'streams' configuration is present, this method:
        1. Calls source.set_streams_config() to let the source enrich its own config
        2. Builds destination record_schemas from streams config
        3. Injects record_schemas into destination config for backward compatibility

        The source is responsible for modifying self.config (which points to bizon_config.source)
        so that subsequent source instantiations see the enriched config.
        """
        if not self.bizon_config.streams:
            return

        logger.info(f"Applying streams configuration: {len(self.bizon_config.streams)} streams defined")

        # Let the source enrich its own config from streams
        # Note: source modifies self.config, which is a reference to bizon_config.source
        # This ensures init_job (which creates a new source) sees the enriched config
        if source and hasattr(source, "set_streams_config") and callable(source.set_streams_config):
            source.set_streams_config(self.bizon_config.streams)

        # Build record_schemas list for destination from streams config
        record_schemas = []
        for stream in self.bizon_config.streams:
            if stream.destination.record_schema:
                record_schema_config = BigQueryRecordSchemaConfig(
                    destination_id=stream.destination.table_id,
                    record_schema=stream.destination.record_schema,
                    clustering_keys=stream.destination.clustering_keys,
                )
                record_schemas.append(record_schema_config)
                logger.info(
                    f"Stream '{stream.name}': "
                    f"{getattr(stream.source, 'topic', getattr(stream.source, 'name', 'N/A'))} "
                    f"-> {stream.destination.table_id}"
                )

        # Inject into destination config
        if record_schemas and hasattr(self.bizon_config.destination.config, "record_schemas"):
            logger.info(f"Injecting {len(record_schemas)} record schemas into destination config")
            self.bizon_config.destination.config.record_schemas = record_schemas

    def run(self) -> RunnerStatus:
        # Create a temporary source to enrich bizon_config.source from streams
        # The source's set_streams_config() modifies self.config (= bizon_config.source)
        # This ensures subsequent source instantiations see the enriched config
        temp_source = self.get_source(bizon_config=self.bizon_config, config=self.config)
        self._apply_streams_config(temp_source)

        # Now initialize job (check_connection will use enriched source config)
        job = self.init_job(bizon_config=self.bizon_config, config=self.config)
        backend = self.get_backend(bizon_config=self.bizon_config)
        source = self.get_source(bizon_config=self.bizon_config, config=self.config)

        sync_metadata = SyncMetadata.from_bizon_config(job_id=job.id, config=self.bizon_config)
        monitor = self.get_monitoring_client(sync_metadata=sync_metadata, bizon_config=self.bizon_config)

        destination = self.get_destination(
            bizon_config=self.bizon_config,
            backend=backend,
            job_id=job.id,
            source_callback=None,
            monitor=monitor,
        )

        transform = self.get_transform(bizon_config=self.bizon_config)

        destination.buffer.buffer_size = 0  # force buffer to be flushed immediately
        iteration = 0

        while True:
            if source.config.max_iterations and iteration > source.config.max_iterations:
                logger.info(f"Max iterations {source.config.max_iterations} reached, terminating stream ...")
                break

            with monitor.trace(operation_name="bizon.stream.iteration"):
                source_iteration = source.get()

                destination_id_indexed_records = {}

                if len(source_iteration.records) == 0:
                    logger.info("No new records found, stopping iteration")
                    time.sleep(2)
                    monitor.track_pipeline_status(PipelineReturnStatus.SUCCESS)
                    iteration += 1
                    continue

                for record in source_iteration.records:
                    if destination_id_indexed_records.get(record.destination_id):
                        destination_id_indexed_records[record.destination_id].append(record)
                    else:
                        destination_id_indexed_records[record.destination_id] = [record]

                for destination_id, records in destination_id_indexed_records.items():
                    df_source_records = StreamingRunner.convert_source_records(records)

                    dsm_headers = monitor.track_source_iteration(records=records)

                    # Apply transformation
                    df_source_records = transform.apply_transforms(df_source_records=df_source_records)

                    df_destination_records = StreamingRunner.convert_to_destination_records(
                        df_source_records, datetime.now(tz=UTC)
                    )
                    # Override destination_id
                    destination.destination_id = destination_id
                    destination.write_or_buffer_records(
                        df_destination_records=df_destination_records,
                        iteration=iteration,
                        pagination=None,
                    )
                    monitor.track_records_synced(
                        num_records=len(df_destination_records),
                        destination_id=destination_id,
                        extra_tags={"destination_id": destination_id},
                        headers=dsm_headers,
                    )

                if os.getenv("ENVIRONMENT") == "production":
                    try:
                        source.commit()
                    except Exception as e:
                        logger.error(f"Error committing source: {e}")
                        monitor.track_pipeline_status(PipelineReturnStatus.ERROR)
                        return RunnerStatus(stream=PipelineReturnStatus.ERROR)

                iteration += 1

                monitor.track_pipeline_status(PipelineReturnStatus.SUCCESS)

        return RunnerStatus(stream=PipelineReturnStatus.SUCCESS)  # return when max iterations is reached



import os
from contextlib import contextmanager
from typing import Dict, List, Union

from datadog import initialize, statsd
from loguru import logger

from bizon.common.models import SyncMetadata
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.monitoring.config import MonitoringConfig
from bizon.monitoring.monitor import AbstractMonitor
from bizon.source.models import SourceRecord


class DatadogMonitor(AbstractMonitor):
    def __init__(self, sync_metadata: SyncMetadata, monitoring_config: MonitoringConfig):
        super().__init__(sync_metadata, monitoring_config)

        # In Kubernetes, set the host dynamically
        try:
            datadog_host_from_env_var = os.getenv(monitoring_config.config.datadog_host_env_var)
            if datadog_host_from_env_var:
                initialize(
                    statsd_host=datadog_host_from_env_var,
                    statsd_port=monitoring_config.config.datadog_agent_port,
                )
            else:
                initialize(
                    statsd_host=monitoring_config.config.datadog_agent_host,
                    statsd_port=monitoring_config.config.datadog_agent_port,
                )
        except Exception as e:
            logger.info(f"Failed to initialize Datadog agent: {e}")

        self.pipeline_monitor_status = "bizon_pipeline.status"
        self.tags = [
            f"pipeline_name:{self.sync_metadata.name}",
            f"pipeline_stream:{self.sync_metadata.stream_name}",
            f"pipeline_source:{self.sync_metadata.source_name}",
            f"pipeline_destination:{self.sync_metadata.destination_name}",
        ] + [f"{key}:{value}" for key, value in self.monitoring_config.config.tags.items()]

        self.pipeline_active_pipelines = "bizon_pipeline.active_pipelines"
        self.pipeline_records_synced = "bizon_pipeline.records_synced"
        self.pipeline_large_records = "bizon_pipeline.large_records"

    def track_pipeline_status(self, pipeline_status: PipelineReturnStatus, extra_tags: Dict[str, str] = {}) -> None:
        """
        Track the status of the pipeline.

        Args:
            status (str): The current status of the pipeline (e.g., 'running', 'failed', 'completed').
        """

        statsd.increment(
            self.pipeline_monitor_status,
            tags=self.tags
            + [f"pipeline_status:{pipeline_status}"]
            + [f"{key}:{value}" for key, value in extra_tags.items()],
        )

    def track_records_synced(
        self, num_records: int, destination_id: str, extra_tags: Dict[str, str] = {}, headers: List[Dict[str, str]] = []
    ) -> Union[List[Dict[str, str]], None]:
        """
        Track the number of records synced in the pipeline.

        Args:
            num_records (int): Number of records synced in this batch
        """
        statsd.increment(
            self.pipeline_records_synced,
            value=num_records,
            tags=self.tags + [f"{key}:{value}" for key, value in extra_tags.items()],
        )
        if os.getenv("DD_DATA_STREAMS_ENABLED") == "true":
            from ddtrace.data_streams import set_produce_checkpoint

            destination_type = self.sync_metadata.destination_alias

            for header in headers:
                if "x-datadog-sampling-priority" in header:
                    del header["x-datadog-sampling-priority"]
                if "dd-pathway-ctx-base64" in header:
                    del header["dd-pathway-ctx-base64"]
                set_produce_checkpoint(destination_type, destination_id, header.setdefault)
            return headers

    def track_large_records_synced(self, num_records: int, extra_tags: Dict[str, str] = {}) -> None:
        statsd.increment(
            self.pipeline_large_records,
            value=num_records,
            tags=self.tags + [f"{key}:{value}" for key, value in extra_tags.items()],
        )

    def track_source_iteration(self, records: List[SourceRecord]) -> Union[List[Dict[str, str]], None]:
        """
        Track the number of records consumed from a Kafka topic.

        Args:
            kafka_topic (str): The Kafka topic name
        """

        if os.getenv("DD_DATA_STREAMS_ENABLED") == "true":
            from ddtrace.data_streams import set_consume_checkpoint

            headers_list = []
            for record in records:
                headers = record.data.get("headers", {})
                set_consume_checkpoint("kafka", record.data["topic"], headers.get)
                headers_list.append(headers)
            return headers_list

    @contextmanager
    def trace(self, operation_name: str, resource: str = None, extra_tags: Dict[str, str] = None):
        """
        Create a trace span for monitoring using Datadog APM.

        Args:
            operation_name (str): The name of the operation being traced
            resource (str): The resource being operated on (e.g., topic name, table name)
            extra_tags (Dict[str, str]): Additional tags for the trace

        Yields:
            A span object that can be used to add additional metadata
        """
        if not self.monitoring_config.config.enable_tracing:
            yield None
            return

        try:
            from ddtrace import tracer
        except ImportError:
            logger.warning("ddtrace not available, skipping tracing")
            yield None
            return

        try:
            # Combine tags
            all_tags = self.tags.copy()
            if extra_tags:
                all_tags.extend([f"{key}:{value}" for key, value in extra_tags.items()])

            # Create the span
            with tracer.trace(operation_name, resource=resource) as span:
                # Add tags to the span
                for tag in all_tags:
                    if ":" in tag:
                        key, value = tag.split(":", 1)
                        span.set_tag(key, value)
                span.set_tag("_sampling_priority_v1", 1)
                yield span
        except Exception as e:
            logger.warning(f"Failed to create trace: {e}")
            yield None



from enum import Enum
from typing import Any, Dict, List

from pydantic import Field

from bizon.source.config import SourceConfig


class NotionStreams(str, Enum):
    DATABASES = "databases"
    DATA_SOURCES = "data_sources"
    PAGES = "pages"
    BLOCKS = "blocks"
    BLOCKS_MARKDOWN = "blocks_markdown"
    USERS = "users"
    # Streams that fetch all accessible content (no database_ids/page_ids required)
    ALL_PAGES = "all_pages"
    ALL_DATABASES = "all_databases"
    ALL_DATA_SOURCES = "all_data_sources"
    ALL_BLOCKS_MARKDOWN = "all_blocks_markdown"


class NotionSourceConfig(SourceConfig):
    stream: NotionStreams

    database_ids: List[str] = Field(
        default_factory=list,
        description="List of Notion database IDs to fetch. Required for databases, data_sources, pages, and blocks streams.",
    )
    page_ids: List[str] = Field(
        default_factory=list,
        description="List of Notion page IDs to fetch. Used for pages and blocks streams.",
    )
    fetch_blocks_recursively: bool = Field(
        default=True,
        description="Whether to fetch nested blocks recursively. Only applies to blocks stream.",
    )
    max_recursion_depth: int = Field(
        default=5,
        ge=1,
        le=100,
        description="Maximum nesting depth for recursive block fetching. Prevents infinite loops.",
    )
    page_size: int = Field(
        default=100,
        ge=1,
        le=100,
        description="Number of results per page (max 100)",
    )
    max_workers: int = Field(
        default=3,
        ge=1,
        le=10,
        description="Number of concurrent workers for fetching blocks. Keep low to respect rate limits.",
    )
    database_filters: Dict[str, Any] = Field(
        default_factory=dict,
        description="Map of database_id -> Notion filter object. Filters are passed directly to Notion API.",
    )



from enum import Enum


class PipelineReturnStatus(str, Enum):
    """Producer error types"""

    BACKEND_ERROR = "backend_error"
    DESTINATION_ERROR = "destination_error"
    KILLED_BY_RUNNER = "killed_by_runner"
    QUEUE_ERROR = "queue_error"
    RUNNING = "running"
    SOURCE_ERROR = "source_error"
    SUCCESS = "success"
    TRANSFORM_ERROR = "transform_error"
    STREAM_ERROR = "stream_error"



import io
import json
import os
import struct
from unittest.mock import patch

import fastavro
import pytest
from avro.schema import parse

from bizon.connectors.sources.kafka.src.decode import (
    Hashabledict,
    decode_avro_message,
    parse_global_id_from_serialized_message,
)


@pytest.fixture
def test_data():
    # Load test data
    current_dir = os.path.dirname(os.path.abspath(__file__))

    with open(os.path.join(current_dir, "schema.json")) as f:
        schema_dict = json.load(f)

    with open(os.path.join(current_dir, "data.json")) as f:
        data = json.load(f)

    # Create a hashable dictionary from the schema
    hashable_schema = Hashabledict(schema_dict)

    # Parse schema for Avro
    avro_schema = parse(json.dumps(schema_dict))

    # Create a serialized message for testing
    # 1-byte magic byte + 8-byte schema ID + Avro data
    schema_id = 12345

    # Create Avro binary data
    avro_binary = io.BytesIO()
    fastavro.schemaless_writer(avro_binary, avro_schema.to_json(), data)
    avro_data = avro_binary.getvalue()

    # Create mock serialized message with 8-byte schema ID
    magic_byte = struct.pack("b", 0)
    schema_id_bytes = struct.pack(">q", schema_id)
    serialized_message_8 = magic_byte + schema_id_bytes + avro_data

    # Create mock serialized message with 4-byte schema ID
    magic_byte = struct.pack("b", 0)
    schema_id_bytes_4 = struct.pack(">I", schema_id)
    serialized_message_4 = magic_byte + schema_id_bytes_4 + avro_data

    return {
        "schema_dict": schema_dict,
        "data": data,
        "hashable_schema": hashable_schema,
        "avro_schema": avro_schema,
        "schema_id": schema_id,
        "serialized_message_8": serialized_message_8,
        "serialized_message_4": serialized_message_4,
    }


def test_parse_global_id_8_bytes(test_data):
    """Test parsing global ID from message with 8-byte schema ID"""
    # Parse directly from the full message (Apicurio format with 0 as schema ID triggers 8-byte read)
    schema_id, nb_bytes = parse_global_id_from_serialized_message(test_data["serialized_message_8"][:9])
    assert nb_bytes == 8


def test_parse_global_id_4_bytes(test_data):
    """Test parsing global ID from message with 4-byte schema ID"""
    # Parse directly from the full message (Confluent format)
    schema_id, nb_bytes = parse_global_id_from_serialized_message(test_data["serialized_message_4"])
    assert schema_id == test_data["schema_id"]
    assert nb_bytes == 4


def test_parse_global_id_invalid_message():
    """Test parsing global ID with invalid message"""
    from confluent_kafka.serialization import SerializationError

    with pytest.raises(SerializationError):
        # Too short message
        parse_global_id_from_serialized_message(b"123")


@patch("bizon.connectors.sources.kafka.src.decode.fastavro.schemaless_reader")
def test_decode_avro_message(mock_reader, test_data):
    """Test decoding an Avro message"""
    # Set up the mock to return our test data
    mock_reader.return_value = test_data["data"]

    # Call the decode function with raw bytes
    result = decode_avro_message(test_data["serialized_message_8"], 8, test_data["avro_schema"])

    # Verify the result
    assert result == test_data["data"]

    # Verify that the reader was called correctly
    mock_reader.assert_called_once()



from abc import ABC, abstractmethod
from typing import Dict, List, Union

from bizon.common.models import SyncMetadata
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.monitoring.config import MonitoringConfig, MonitorType
from bizon.source.models import SourceRecord


class AbstractMonitor(ABC):
    def __init__(self, sync_metadata: SyncMetadata, monitoring_config: MonitoringConfig):
        self.sync_metadata = sync_metadata
        self.monitoring_config = monitoring_config

    @abstractmethod
    def track_pipeline_status(self, pipeline_status: PipelineReturnStatus, extra_tags: Dict[str, str] = {}) -> None:
        """
        Track the status of the pipeline.

        Args:
            status (str): The current status of the pipeline (e.g., 'running', 'failed', 'completed').
        """
        pass

    def track_source_iteration(self, records: List[SourceRecord], headers: Dict[str, str] = {}) -> None:
        """
        Run a process that tracks the source iteration.
        """
        pass

    def track_records_synced(
        self, num_records: int, destination_id: str, extra_tags: Dict[str, str] = {}, headers: Dict[str, str] = {}
    ) -> None:
        """
        Track the number of records synced in the pipeline.
        """
        pass

    def trace(self, operation_name: str, resource: str = None, extra_tags: Dict[str, str] = None):
        """
        Create a trace span for monitoring.

        Args:
            operation_name (str): The name of the operation being traced
            resource (str): The resource being operated on (e.g., topic name, table name)
            extra_tags (Dict[str, str]): Additional tags for the trace

        Returns:
            A context manager that can be used with 'with' statement
        """
        pass

    def track_large_records_synced(self, num_records: int, extra_tags: Dict[str, str] = {}) -> None:
        """
        Track the number of large records synced in the destination system. This aims at helping to identify the source of the large records.
        """
        pass


class MonitorFactory:
    @staticmethod
    def get_monitor(sync_metadata: SyncMetadata, monitoring_config: Union[MonitoringConfig, None]) -> AbstractMonitor:
        if monitoring_config is None:
            from bizon.monitoring.noop.monitor import NoOpMonitor

            return NoOpMonitor(sync_metadata, monitoring_config)

        if monitoring_config.type == MonitorType.DATADOG:
            from bizon.monitoring.datadog.monitor import DatadogMonitor

            return DatadogMonitor(sync_metadata, monitoring_config)



import os
import tempfile
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from queue import Queue

import pytest
import yaml
from loguru import logger

from bizon.alerting.models import LogLevel
from bizon.alerting.slack.config import SlackConfig
from bizon.alerting.slack.handler import SlackHandler
from bizon.engine.engine import RunnerFactory


class DummyWebhookHandler(BaseHTTPRequestHandler):
    # Shared queue to store payloads
    payload_queue = Queue()

    def do_POST(self):
        # Read and store the payload
        content_length = int(self.headers["Content-Length"])
        post_data = self.rfile.read(content_length)
        self.payload_queue.put(post_data.decode("utf-8"))

        # Send a response back
        self.send_response(200)
        self.end_headers()
        self.wfile.write(b"OK")


# Subclass HTTPServer to set allow_reuse_address
class ReusableHTTPServer(HTTPServer):
    allow_reuse_address = True


# Function to start the server in a separate thread
def start_dummy_server(host="localhost", port=8123):
    server = ReusableHTTPServer((host, port), DummyWebhookHandler)
    server_thread = threading.Thread(target=server.serve_forever, daemon=True)
    server_thread.start()
    return server, server_thread


@pytest.fixture
def dummy_webhook_server():
    # Start the dummy server
    server, server_thread = start_dummy_server()

    # Yield control to the test
    yield server

    # Shutdown the server after the test
    server.shutdown()
    server_thread.join()

    # Clear the payload queue
    DummyWebhookHandler.payload_queue.queue.clear()


@pytest.fixture
def webhook_url():
    return "http://localhost:8123"


def test_slack_log_handler(dummy_webhook_server, webhook_url):
    slack_handler = SlackHandler(SlackConfig(webhook_url=webhook_url), log_levels=[LogLevel.ERROR, LogLevel.WARNING])

    slack_handler.add_handlers()

    ERROR_MESSAGE = "This is an error message"
    WARNING_MESSAGE = "This is a warning message"

    logger.error(ERROR_MESSAGE)

    error_payload = DummyWebhookHandler.payload_queue.get(timeout=1)
    assert ERROR_MESSAGE in error_payload

    DummyWebhookHandler.payload_queue.queue.clear()

    logger.warning(WARNING_MESSAGE)
    warning_payload = DummyWebhookHandler.payload_queue.get(timeout=1)
    assert WARNING_MESSAGE in warning_payload


def test_e2e_logger_to_file(dummy_webhook_server, webhook_url):
    with tempfile.NamedTemporaryFile(delete=False) as temp:
        BIZON_CONFIG_DUMMY_TO_FILE = f"""
        name: test_job_9

        source:
          name: dummy
          stream: creatures
          authentication:
            type: api_key
            params:
              token: dummy_key

        destination:
          name: file
          config:
            destination_id: {temp.name}

        transforms:
          - label: transform_data
            python: |
              if 'name' in data:
                data['name'] = fake_variable # this is purposely wrong to trigger an error

        engine:
          backend:
            type: postgres
            config:
              database: bizon_test
              schema: public
              syncCursorInDBEvery: 2
              host: {os.environ.get("POSTGRES_HOST", "localhost")}
              port: 5432
              username: postgres
              password: bizon

        alerting:
            type: slack

            config:
              webhook_url: {webhook_url}

            log_levels:
                - ERROR
        """

        runner = RunnerFactory.create_from_config_dict(yaml.safe_load(BIZON_CONFIG_DUMMY_TO_FILE))

        runner.run()

        error_payload = DummyWebhookHandler.payload_queue.get(timeout=1)
        assert "Error applying transformation" in error_payload
        assert "fake_variable" in error_payload



from datetime import datetime

import polars as pl
import pytest

from bizon.common.models import SyncMetadata
from bizon.connectors.destinations.logger.src.config import LoggerDestinationConfig
from bizon.connectors.destinations.logger.src.destination import LoggerDestination
from bizon.destination.destination import DestinationBufferStatus
from bizon.destination.models import destination_record_schema
from bizon.engine.backend.adapters.sqlalchemy.backend import SQLAlchemyBackend
from bizon.engine.backend.models import JobStatus, StreamJob
from bizon.monitoring.noop.monitor import NoOpMonitor
from bizon.source.callback import NoOpSourceCallback


@pytest.fixture(scope="function")
def logger_destination(my_sqlite_backend: SQLAlchemyBackend, sqlite_db_session):
    my_sqlite_backend.create_all_tables()

    job = my_sqlite_backend.create_stream_job(
        name="job_test",
        source_name="dummy",
        stream_name="test",
        sync_mode="full_refresh",
        job_status=JobStatus.RUNNING,
        session=sqlite_db_session,
    )

    sync_metadata = SyncMetadata(
        job_id=job.id,
        name="job_test",
        source_name="dummy",
        stream_name="test",
        destination_name="logger",
        destination_alias="logger",
        sync_mode="full_refresh",
    )

    return LoggerDestination(
        sync_metadata=sync_metadata,
        config=LoggerDestinationConfig(dummy="bizon"),
        backend=my_sqlite_backend,
        source_callback=NoOpSourceCallback(config={}),
        monitor=NoOpMonitor(sync_metadata=sync_metadata, monitoring_config=None),
    )


df_destination_records = pl.DataFrame(
    {
        "bizon_id": ["id_1", "id_2"],
        "bizon_extracted_at": [datetime(2024, 12, 5, 12, 0), datetime(2024, 12, 5, 13, 0)],
        "bizon_loaded_at": [datetime(2024, 12, 5, 12, 30), datetime(2024, 12, 5, 13, 30)],
        "source_record_id": ["record_1", "record_2"],
        "source_timestamp": [datetime(2024, 12, 5, 11, 30), datetime(2024, 12, 5, 12, 30)],
        "source_data": ["cookies", "cream"],
    },
    schema=destination_record_schema,
)


def test_logger_destination(logger_destination: LoggerDestination):
    assert logger_destination.buffer.df_destination_records.height == 0


def test_buffer_records(logger_destination: LoggerDestination):
    logger_destination.buffer.add_source_iteration_records_to_buffer(
        iteration=0, df_destination_records=df_destination_records
    )
    assert logger_destination.buffer.df_destination_records.equals(df_destination_records)


def test_write_or_buffer_records_too_large(logger_destination: LoggerDestination):
    df_big_size = pl.DataFrame(schema=destination_record_schema)

    for _ in range(100):
        df_big_size = df_big_size.vstack(df_destination_records, in_place=True)

    # Overriding buffer size to bytes instead of mbs
    logger_destination.buffer.buffer_size = df_big_size.estimated_size(unit="b")

    buffer_status = logger_destination.write_or_buffer_records(
        df_destination_records=df_destination_records,
        iteration=0,
    )

    assert buffer_status == DestinationBufferStatus.RECORDS_BUFFERED

    # Reset buffer
    logger_destination.buffer.buffer_size = df_big_size.estimated_size(unit="b")

    # Write twice
    with pytest.raises(
        ValueError, match="Please increase destination buffer_size or reduce batch_size from the source"
    ):
        buffer_status = logger_destination.write_or_buffer_records(
            df_destination_records=df_big_size.vstack(df_destination_records), iteration=1
        )


def test_write_last_iteration(logger_destination: LoggerDestination, sqlite_db_session):
    buffer_status = logger_destination.write_or_buffer_records(
        df_destination_records=df_destination_records,
        iteration=0,
        session=sqlite_db_session,
    )

    assert buffer_status == DestinationBufferStatus.RECORDS_BUFFERED

    running_job: StreamJob = logger_destination.backend.get_stream_job_by_id(
        job_id=logger_destination.sync_metadata.job_id, session=sqlite_db_session
    )
    assert running_job.status == JobStatus.RUNNING

    # Write the last iteration
    buffer_status = logger_destination.write_or_buffer_records(
        df_destination_records=pl.DataFrame(schema=destination_record_schema),
        iteration=1,
        last_iteration=True,
        session=sqlite_db_session,
    )

    assert buffer_status == DestinationBufferStatus.RECORDS_WRITTEN

    finished_job: StreamJob = logger_destination.backend.get_stream_job_by_id(
        job_id=logger_destination.sync_metadata.job_id, session=sqlite_db_session
    )
    assert finished_job.status == JobStatus.SUCCEEDED


def test_no_records_written(logger_destination: LoggerDestination, sqlite_db_session):
    buffer_status = logger_destination.write_or_buffer_records(
        df_destination_records=pl.DataFrame(schema=destination_record_schema),
        iteration=0,
        session=sqlite_db_session,
    )

    assert buffer_status == DestinationBufferStatus.NO_RECORDS



import ast
import importlib
import importlib.util
import inspect
import os
import traceback
from collections.abc import Mapping
from typing import Any, List, Type

from loguru import logger
from pydantic import BaseModel

from bizon.source.source import AbstractSource
from bizon.utils import BIZON_ABSOLUTE_PATH


class Stream(BaseModel):
    name: str
    source_class: Type[AbstractSource]
    supports_incremental: bool


class SourceModel(BaseModel):
    name: str
    streams: List[Stream]

    @property
    def available_streams(self) -> List[str]:
        return [stream.name for stream in self.streams]

    def get_stream_by_name(self, name: str) -> Stream:
        for stream in self.streams:
            if stream.name == name:
                return stream
        raise ValueError(f"Stream {name} not found. Available streams are {self.available_streams}")


def find_inherited_classes(file_path):
    # Open the file and parse its content using ast
    with open(file_path) as file:
        tree = ast.parse(file.read())

    # List to store classes that inherit from the given class name
    inherited_classes = []

    # Walk through the AST nodes
    for node in ast.walk(tree):
        # Check if the node is a class definition
        if isinstance(node, ast.ClassDef):
            # Check if the class inherits from the parent_class_name
            for base in node.bases:
                if isinstance(base, ast.Name):
                    inherited_classes.append(node.name)

    return inherited_classes


def get_python_import_path(relative_path: str) -> str:
    """Transform a relative path to a python import path"""
    python_path = relative_path.replace("/", ".").replace(".py", "")

    # Find the index where the desired substring starts
    start_index = python_path.find("bizon.connectors.sources")

    # Extract the substring from that point onwards
    if start_index != -1:
        return python_path[start_index:]
    raise ValueError("Substring not found.")


def find_all_source_paths(source_dir_name: str = "src") -> Mapping[str, str]:
    """Find all source code paths in the sources directory
    Return a dict mapping source_name to source_code_path
    Like {'gsheets': '/path/to/bizon/connectors/sources/gsheets/src'}
    """

    discovered_sources_paths = {}

    # Leading to the bizon internal sources directory
    base_dir = os.path.join(BIZON_ABSOLUTE_PATH, "connectors", "sources")

    for source_name in os.listdir(base_dir):
        # First check that it's a dir and contains a dir called src
        if os.path.isdir(os.path.join(base_dir, source_name)) and os.path.exists(
            os.path.join(base_dir, source_name, source_dir_name)
        ):
            # Store path to source code
            source_code_path = os.path.join(base_dir, source_name, source_dir_name)
            discovered_sources_paths[source_name] = source_code_path

    return discovered_sources_paths


def is_class_a_source(source_class: Type, path: str) -> bool:
    """Check if a class is a source"""

    source_class_name = source_class.__name__

    # Get the source class from the module
    if inspect.isabstract(source_class):
        logger.debug(f"Class {source_class_name} in {path} is abstract. Skipping.")
        return False

    if not issubclass(source_class, AbstractSource):
        logger.debug(f"Class {source_class_name} in {path} does not inherit from AbstractSource. Skipping.")
        return False

    # Check if the class has a streams method
    if not hasattr(source_class, "streams") or not callable(source_class.streams):
        logger.warning(
            f"Class {source_class_name} in {path} does not have a streams method."
            "Please add a streams method to the class to be able to use it."
        )
        return False

    if len(source_class.streams()) == 0:
        logger.warning(
            f"Class {source_class_name} in {path} has no streams defined."
            "Please add at least one stream to the class to be able to use it."
        )
        return False

    return True


def is_source_class_implementing_incremental(source_class: Type[AbstractSource]) -> bool:
    """Check if a source class implements incremental"""
    source_code = inspect.getsource(source_class.get_records_after).strip()
    supports_incremental = not source_code.endswith("pass")
    return supports_incremental


def parse_streams_from_filepath(source_name: str, filepath: str, skip_unavailable_sources: bool) -> List[Stream]:
    streams = []

    # Find all classes that inherit from AbstractSource
    source_classes_name = find_inherited_classes(file_path=filepath)

    relative_path: str = os.path.relpath(filepath)

    # If classes are found
    for source_class_name in source_classes_name:
        # Transform the relative path to a python import path and import the module
        python_import_path = get_python_import_path(relative_path)
        logger.debug(f"Importing {python_import_path}")

        try:
            source_module = importlib.import_module(python_import_path, package="sources")
        except ImportError as e:
            if not skip_unavailable_sources:
                raise e
            else:
                logger.warning(
                    f"{source_name} is not available, run 'pip install bizon[{source_name}]' to install missing dependencies."
                )
            break
        except Exception as e:
            logger.error(f"Error while importing {python_import_path}: {e}")
            logger.error(traceback.format_exc())
            break

        source_class = getattr(source_module, source_class_name)

        # Get the source class from the module
        if is_class_a_source(source_class=source_class, path=relative_path):
            for stream in source_class.streams():
                streams.append(
                    Stream(
                        name=stream,
                        source_class=source_class,
                        supports_incremental=is_source_class_implementing_incremental(source_class),
                    )
                )

    return streams


def get_sources_in_path(source_name: str, source_code_path: str, skip_unavailable_sources: bool = True) -> SourceModel:
    """Get all sources in a given path
    Return a SourceModel object with the source name and its streams
    """

    streams = []

    # Iterate all files in the source code path recursively
    for root, _, files in os.walk(source_code_path):
        for file in files:
            if file.endswith(".py"):
                streams.extend(
                    parse_streams_from_filepath(
                        source_name=source_name,
                        filepath=os.path.join(root, file),
                        skip_unavailable_sources=skip_unavailable_sources,
                    )
                )

    return SourceModel(name=source_name, streams=streams)


def get_internal_source_class_by_source_and_stream(source_name: str, stream_name: str) -> Type[AbstractSource]:
    """Get the source class by source and stream name"""

    discovered_sources_paths_dict = find_all_source_paths()

    if source_name not in discovered_sources_paths_dict.keys():
        raise ValueError(
            f"Source {source_name} not found. Available sources are {discovered_sources_paths_dict.keys()}"
        )

    source_code_path = discovered_sources_paths_dict[source_name]

    source_model = get_sources_in_path(
        source_name=source_name, source_code_path=source_code_path, skip_unavailable_sources=False
    )

    if len(source_model.streams) == 0:
        raise ValueError(
            f"No streams found for source {source_name}, ensure you installed required dependencies with pip install bizon[{source_name}] if needed."
            "You should not have any warning related to your source when running 'bizon source list'."
        )

    if stream_name not in source_model.available_streams:
        raise ValueError(f"Stream {stream_name} not found. Available streams are {source_model.available_streams}")

    stream = source_model.get_stream_by_name(name=stream_name)

    return stream.source_class


def get_external_source_class_by_source_and_stream(
    source_name: str, stream_name: str, filepath: str
) -> Type[AbstractSource]:
    """Get the source class by source and stream name"""

    if not os.path.exists(filepath):
        raise ValueError(f"File {filepath} not found.")

    if not filepath.endswith(".py"):
        raise ValueError(f"File {filepath} is not a python file. File extension should be '.py'")

    # Find  class that inherit from AbstractSource
    class_names = find_inherited_classes(file_path=filepath)

    stream_names = []

    for class_name in class_names:
        # Extract the module name from the file (remove directory and extension)
        module_name = os.path.splitext(os.path.basename(filepath))[0]

        # Create module spec from the file
        spec = importlib.util.spec_from_file_location(module_name, filepath)

        # Create a module from the spec
        module = importlib.util.module_from_spec(spec)

        # Execute the module (this loads the module into memory)
        spec.loader.exec_module(module)

        class_to_test = getattr(module, class_name)

        if is_class_a_source(class_to_test, filepath):
            if stream_name in class_to_test.streams():
                return class_to_test

            stream_names.extend(class_to_test.streams())

    raise ValueError(f"Stream {stream_name} not found. Available streams are {stream_names}")


def discover_all_sources() -> Mapping[str, SourceModel]:
    """Discover all sources in the sources directory
    Return a dict mapping source_name to SourceModel object
    """

    discovered_sources = {}

    discovered_sources_paths = find_all_source_paths()

    # Iterate all source code paths found
    for source_name, source_code_path in discovered_sources_paths.items():
        try:
            source_model = get_sources_in_path(source_name=source_name, source_code_path=source_code_path)
            discovered_sources[source_name] = source_model
        except Exception as e:
            logger.error(f"Error while discovering source {source_name}: {e}")
            logger.error(traceback.format_exc())

    return discovered_sources


def get_source_instance_by_source_and_stream(
    source_name: str, stream_name: str, source_config: Mapping[str, Any]
) -> AbstractSource:
    """Get an instance of the source by source and stream name"""

    if source_config.get("source_file_path"):
        source_class: AbstractSource = get_external_source_class_by_source_and_stream(
            source_name=source_name,
            stream_name=stream_name,
            filepath=source_config["source_file_path"],
        )
    else:
        source_class: AbstractSource = get_internal_source_class_by_source_and_stream(
            source_name=source_name, stream_name=stream_name
        )

    config_class = source_class.get_config_class()
    config_parsed = config_class.model_validate(source_config)

    return source_class(config=config_parsed)



from os import getenv

from bizon.cli.utils import parse_from_yaml
from bizon.common.models import BizonConfig

from .config import RunnerTypes
from .runner.runner import AbstractRunner


def replace_env_variables_in_config(config: dict) -> dict:
    """Replace templated secrets with actual values from environment variables"""
    for key, value in config.items():
        if isinstance(value, dict):
            config[key] = replace_env_variables_in_config(value)
        elif isinstance(value, str):
            if value.startswith("BIZON_ENV_"):
                config[key] = getenv(value)
    return config


class RunnerFactory:
    @staticmethod
    def create_from_config_dict(config: dict) -> AbstractRunner:
        # Replace env variables in config
        config = replace_env_variables_in_config(config=config)

        bizon_config = BizonConfig.model_validate(obj=config)

        if bizon_config.engine.runner.type == RunnerTypes.THREAD:
            from .runner.adapters.thread import ThreadRunner

            return ThreadRunner(config=config)

        if bizon_config.engine.runner.type == RunnerTypes.PROCESS:
            from .runner.adapters.process import ProcessRunner

            return ProcessRunner(config=config)

        if bizon_config.engine.runner.type == RunnerTypes.STREAM:
            from .runner.adapters.streaming import StreamingRunner

            return StreamingRunner(config=config)

        raise ValueError(f"Runner type {bizon_config.engine.runner.type} is not supported")

    @staticmethod
    def create_from_yaml(filepath: str) -> AbstractRunner:
        config = parse_from_yaml(filepath)
        return RunnerFactory.create_from_config_dict(config)



from bizon.connectors.sources.hubspot.src.hubspot_objects import (
    URL_TOKEN_REFRESH,
    HubSpotObjectsSource,
    HubSpotSourceConfig,
)
from bizon.source.auth.config import AuthType


def test_source_config_valid():
    raw_config = {
        "source": {
            "name": "hubspot",
            "stream": "contacts",
            "authentication": {
                "type": "oauth",
                "params": {
                    "token_refresh_endpoint": "https://api.hubapi.com/oauth/v1/token",
                    "client_id": "id-12345abcd",
                    "client_secret": "cli-12345abcd",
                    "refresh_token": "fresh-12345abcd",
                },
            },
        }
    }
    source_config = HubSpotSourceConfig.model_validate(raw_config["source"])
    assert source_config.authentication.type == AuthType.OAUTH
    assert source_config.authentication.params.client_id == "id-12345abcd"
    assert source_config.authentication.params.client_secret == "cli-12345abcd"
    assert source_config.authentication.params.refresh_token == "fresh-12345abcd"
    assert source_config.authentication.params.token_refresh_endpoint == URL_TOKEN_REFRESH


def test_source_instanciation_contacts():
    raw_config = {
        "source": {
            "name": "hubspot",
            "stream": "contacts",
            "init_pipeline": False,
            "authentication": {
                "type": "oauth",
                "params": {
                    "token_refresh_endpoint": "https://api.hubapi.com/oauth/v1/token",
                    "client_id": "id-12345abcd",
                    "client_secret": "cli-12345abcd",
                    "refresh_token": "fresh-12345abcd",
                },
            },
        }
    }
    client = HubSpotObjectsSource(config=HubSpotSourceConfig.model_validate(raw_config["source"]))
    assert client.object == "contacts"


def test_source_instanciation_companies():
    raw_config = {
        "source": {
            "name": "hubspot",
            "stream": "companies",
            "init_pipeline": False,
            "authentication": {
                "type": "oauth",
                "params": {
                    "token_refresh_endpoint": "https://api.hubapi.com/oauth/v1/token",
                    "client_id": "id-12345abcd",
                    "client_secret": "cli-12345abcd",
                    "refresh_token": "fresh-12345abcd",
                },
            },
        }
    }
    client = HubSpotObjectsSource(config=HubSpotSourceConfig.model_validate(raw_config["source"]))
    assert client.object == "companies"



from bizon.connectors.sources.kafka.src.config import KafkaAuthConfig, TopicConfig
from bizon.connectors.sources.kafka.src.source import KafkaSource, KafkaSourceConfig


def test_kafka_source_config_timestamp_to_parse():
    conf = KafkaSourceConfig(
        name="kafka",
        stream="topic",
        topics=[
            TopicConfig(name="cookie", destination_id="cookie"),
        ],
        bootstrap_servers="fdjvfv",
        batch_size=87,
        consumer_timeout=56,
        authentication=KafkaAuthConfig(
            type="basic",
            params={"username": "user", "password": "password"},
        ),
    )
    assert conf


def test_kafka_source_config_check_connection():
    conf = KafkaSourceConfig(
        name="kafka",
        stream="topic",
        authentication=KafkaAuthConfig(
            type="basic",
            params={"username": "user", "password": "password"},
        ),
        bootstrap_servers="fdjvfv",  # Invalid bootstrap server
        topics=[
            TopicConfig(name="cookie", destination_id="cookie"),
        ],
    )

    source = KafkaSource(config=conf)
    success, error = source.check_connection()

    # Should fail due to invalid bootstrap server and/or credentials
    assert success is False
    assert error is not None
    assert "connection" in error.lower() or "failed" in error.lower()



from pydantic import BaseModel, Field


class TransformModel(BaseModel):
    label: str = Field(..., description="Label of the transform")
    python: str = Field(..., description="Python code for the transform")



from typing import Literal, Optional

from bizon.destination.config import (
    AbstractDestinationConfig,
    AbstractDestinationDetailsConfig,
    DestinationTypes,
)


class LoggerDestinationConfig(AbstractDestinationDetailsConfig):
    dummy: Optional[str] = "bizon"


class LoggerConfig(AbstractDestinationConfig):
    name: Literal[DestinationTypes.LOGGER]
    alias: str = "logger"
    config: LoggerDestinationConfig



import pytest

from bizon.source.discover import (
    discover_all_sources,
    find_all_source_paths,
    get_external_source_class_by_source_and_stream,
    get_internal_source_class_by_source_and_stream,
    get_source_instance_by_source_and_stream,
    parse_streams_from_filepath,
)

DUMMY_SOURCE_CONFIG_DICT = {
    "name": "dummy",
    "stream": "creatures",
    "config": {"dummy": "dummy"},
    "authentication": {"type": "api_key", "params": {"token": "my_dummy_token"}},
}

DUMMY_EXTERNAL_SOURCE_CONFIG_DICT = {
    "name": "dummy",
    "stream": "flowers",
    "config": {"dummy": "dummy"},
    "authentication": {"type": "api_key", "params": {"token": "my_dummy"}},
    "source_file_path": "./tests/source/custom_source.py",
}


def test_find_all_source_paths():
    source_paths = find_all_source_paths()
    assert len(source_paths) > 0
    assert "dummy" in source_paths


def test_parse_streams_from_filepath():
    streams = parse_streams_from_filepath(
        source_name="dummy", filepath="bizon/connectors/sources/dummy/src/source.py", skip_unavailable_sources=True
    )
    assert len(streams) > 0
    set(stream.name for stream in streams) == set(["creatures", "plants"])


def test_get_internal_source_class_by_source_and_stream():
    source_class = get_internal_source_class_by_source_and_stream("dummy", "creatures")
    assert source_class.__name__ == "DummySource"


def test_get_internal_source_class_by_source_and_stream_stream_not_found():
    with pytest.raises(ValueError) as error:
        source_class = get_internal_source_class_by_source_and_stream("dummy", "invalid_stream")
    assert "Stream invalid_stream not found. Available streams are" in str(error.value)


def test_get_internal_source_class_by_source_and_stream_stream_source_invalid():
    with pytest.raises(ValueError) as error:
        source_class = get_internal_source_class_by_source_and_stream("invalid_source", "inexistent_stream")
    assert "Source invalid_source not found. Available sources are" in str(error.value)


def test_get_external_source_class_by_source_and_stream():
    source_class = get_external_source_class_by_source_and_stream(
        source_name="dummy",
        stream_name="flowers",
        filepath="./tests/source/custom_source.py",
    )
    assert source_class.__name__ == "MyDummyCustomSource"


def test_get_external_source_class_by_source_and_stream_invalid_stream():
    with pytest.raises(ValueError) as error:
        source_class = get_external_source_class_by_source_and_stream(
            source_name="dummy",
            stream_name="creatures",
            filepath="./tests/source/custom_source.py",
        )
    assert "Stream creatures not found. Available streams are ['flowers']" in str(error.value)


def test_get_source_instance_by_source_and_stream():
    source_instance = get_source_instance_by_source_and_stream("dummy", "creatures", DUMMY_SOURCE_CONFIG_DICT)
    assert source_instance.__class__.__name__ == "DummySource"


def test_get_source_instance_by_source_and_stream_external_file():
    source_instance = get_source_instance_by_source_and_stream("dummy", "flowers", DUMMY_EXTERNAL_SOURCE_CONFIG_DICT)
    assert source_instance.__class__.__name__ == "MyDummyCustomSource"


def test_discover_all_sources():
    all_sources = discover_all_sources()

    assert len(all_sources) > 0

    found_dummy = False

    for source_name, source_model in all_sources.items():
        if source_name == "dummy":
            found_dummy = True
            assert set(source_model.available_streams) == set(["creatures", "plants"])

    assert found_dummy



import os

import pytest

from bizon.engine.backend.adapters.sqlalchemy.backend import SQLAlchemyBackend
from bizon.engine.backend.adapters.sqlalchemy.config import (
    BigQueryConfigDetails,
    BigQuerySQLAlchemyConfig,
    PostgresConfigDetails,
    PostgresSQLAlchemyConfig,
    SQLiteConfigDetails,
    SQLiteInMemoryConfig,
)
from bizon.engine.backend.backend import BackendFactory
from bizon.engine.backend.config import BackendTypes


def test_backend_factory_sqlite():
    config = SQLiteInMemoryConfig(
        type=BackendTypes.SQLITE_IN_MEMORY,
        config=SQLiteConfigDetails(
            database="NOT_DATABASE_WITH_SQLITE", schema="NOT_SCHEMA_IN_SQLITE", syncCursorInDBEvery=2
        ),
    )

    backend = BackendFactory().get_backend(config=config)
    assert isinstance(backend, SQLAlchemyBackend)


def test_backend_factory_postgres():
    config = PostgresSQLAlchemyConfig(
        type=BackendTypes.POSTGRES,
        config=PostgresConfigDetails(
            database="NOT_DATABASE_WITH_SQLITE",
            schema="NOT_SCHEMA_IN_SQLITE",
            syncCursorInDBEvery=2,
            host="localhost",
            port=5432,
            username="postgres",
            password="bizon",
        ),
    )

    backend = BackendFactory().get_backend(config=config)
    assert isinstance(backend, SQLAlchemyBackend)


@pytest.mark.skipif(
    os.getenv("CI") is not None,
    reason="Skipping tests that require a BigQuery database",
)
def test_backend_factory_bigquery():
    config = BigQuerySQLAlchemyConfig(
        type=BackendTypes.BIGQUERY,
        config=BigQueryConfigDetails(
            database="project",
            schema="dataset",
            syncCursorInDBEvery=2,
        ),
    )

    backend = BackendFactory().get_backend(config=config)
    assert isinstance(backend, SQLAlchemyBackend)



import tempfile

import yaml

from bizon.engine.engine import RunnerFactory
from bizon.engine.pipeline.models import PipelineReturnStatus


def test_e2e_pipeline_should_stop():
    with tempfile.NamedTemporaryFile(delete=False) as temp:
        BIZON_CONFIG_DUMMY_TO_FILE = f"""
        name: test_job_3

        source:
          name: dummy
          stream: creatures

          sleep: 2 # Add sleep to make sure the pipeline is stopped before it finishes

          authentication:
            type: api_key
            params:
              token: dummy_key

        destination:
          name: file
          config:
            destination_id: {temp.name}

        transforms:
          - label: transform_data
            python: |
              if 'name' in data:
                data['name'] = data['this_key_doesnt_exist'].upper()
        """

        runner = RunnerFactory.create_from_config_dict(yaml.safe_load(BIZON_CONFIG_DUMMY_TO_FILE))

        status = runner.run()

    assert status.producer == PipelineReturnStatus.KILLED_BY_RUNNER
    assert status.consumer == PipelineReturnStatus.TRANSFORM_ERROR



from collections.abc import Mapping
from enum import Enum
from typing import Any, List, Literal, Optional

from pydantic import BaseModel, Field

from bizon.source.auth.config import AuthConfig, AuthType
from bizon.source.config import SourceConfig


class SchemaRegistryType(str, Enum):
    APICURIO = "apicurio"


class MessageEncoding(str, Enum):
    UTF_8 = "utf-8"
    AVRO = "avro"


class KafkaAuthConfig(AuthConfig):
    type: Literal[AuthType.BASIC] = AuthType.BASIC  # username and password authentication

    # Schema registry authentication
    schema_registry_type: SchemaRegistryType = Field(
        default=SchemaRegistryType.APICURIO, description="Schema registry type"
    )

    schema_registry_url: str = Field(default="", description="Schema registry URL with the format ")
    schema_registry_username: str = Field(default="", description="Schema registry username")
    schema_registry_password: str = Field(default="", description="Schema registry password")


def default_kafka_consumer_config():
    return {
        "auto.offset.reset": "earliest",
        "enable.auto.commit": False,  # Turn off auto-commit for manual offset handling
        "session.timeout.ms": 45000,
        "security.protocol": "SASL_SSL",
    }


class TopicConfig(BaseModel):
    name: str = Field(..., description="Kafka topic name")
    destination_id: str = Field(..., description="Destination id")


class KafkaSourceConfig(SourceConfig):
    # Kafka configuration
    topics: Optional[List[TopicConfig]] = Field(
        default=[],
        description="Kafka topics. Can be empty if using streams configuration to define topics.",
    )
    bootstrap_servers: str = Field(..., description="Kafka bootstrap servers")
    group_id: str = Field(default="bizon", description="Kafka group id")

    skip_message_empty_value: bool = Field(
        default=True, description="Skip messages with empty value (tombstone messages)"
    )
    skip_message_invalid_keys: bool = Field(
        default=False, description="Skip messages with invalid keys (unparsable JSON keys)"
    )
    # Kafka consumer configuration
    batch_size: int = Field(100, description="Kafka batch size, number of messages to fetch at once.")
    consumer_timeout: int = Field(10, description="Kafka consumer timeout in seconds, before returning batch.")

    consumer_config: Mapping[str, Any] = Field(
        default_factory=default_kafka_consumer_config,
        description="Kafka consumer configuration, as described in the confluent-kafka-python documentation",
    )

    message_encoding: str = Field(default=MessageEncoding.AVRO, description="Encoding to use to decode the message")

    authentication: KafkaAuthConfig = Field(..., description="Authentication configuration")



import json
import os

import yaml

from bizon.engine.engine import RunnerFactory


def test_e2e_dummy_streaming_to_file():
    BIZON_CONFIG_DUMMY_TO_FILE = f"""
      name: test_job_3

      source:
        name: dummy
        stream: creatures
        sync_mode: stream # This will make the source to stream data to the destination
        authentication:
          type: api_key
          params:
            token: dummy_key
        max_iterations: 4

      destination:
        name: file
        config:
          unnest: true
          record_schemas:
            - destination_id: routed
              record_schema:
                - name: id
                  type: string
                  nullable: false
                - name: name
                  type: string
                  nullable: false
                - name: age
                  type: integer
                  nullable: false

            - destination_id: creatures
              record_schema:
                - name: id
                  type: string
                  nullable: false
                - name: name
                  type: string
                  nullable: false
                - name: age
                  type: integer
                  nullable: false

      transforms:
        - label: transform_data
          python: |
            if 'name' in data:
              data['name'] = data['name'].upper()

      engine:
        runner:
          type: stream
        backend:
          type: postgres
          config:
            database: bizon_test
            schema: public
            syncCursorInDBEvery: 2
            host: {os.environ.get("POSTGRES_HOST", "localhost")}
            port: 5432
            username: postgres
            password: bizon
      """

    runner = RunnerFactory.create_from_config_dict(yaml.safe_load(BIZON_CONFIG_DUMMY_TO_FILE))

    runner.run()

    records_extracted = {}
    with open("creatures.json") as file:
        for line in file.readlines():
            record: dict = json.loads(line.strip())
            records_extracted[record["id"]] = record["name"]

    assert set(records_extracted.keys()) == set([9898, 88787])
    assert records_extracted[9898] == "BIZON"



from datetime import datetime
from typing import List, Optional, Union

import polars as pl
from pydantic import BaseModel, Field, field_validator
from pytz import UTC

# Define the SourceRecord model
source_record_schema = pl.Schema(
    [
        ("id", str),
        ("data", str),  # JSON payload dumped as string
        ("timestamp", pl.Datetime(time_unit="us", time_zone="UTC")),
        ("destination_id", str),
    ]
)


### /!\ These models Source* will be used in all sources so we better never have to change them !!!
class SourceRecord(BaseModel):
    id: str = Field(..., description="Unique identifier of the record in the source")

    data: dict = Field(..., description="JSON payload of the record")

    timestamp: datetime = Field(
        default=datetime.now(tz=UTC),
        description="Timestamp of the record as defined by the source. Default is the time of extraction",
    )

    destination_id: Optional[str] = Field(None, description="Destination id")

    @field_validator("id", mode="before")
    def coerce_int_to_str(value: Union[int, str]) -> str:
        # Coerce int to str in case Source return id as int
        if isinstance(value, int):
            return str(value)
        return value


class SourceIteration(BaseModel):
    next_pagination: dict = Field(..., description="Next pagination to be used in the next iteration")
    records: List[SourceRecord] = Field(..., description="List of records retrieved in the current iteration")


class SourceIncrementalState(BaseModel):
    last_run: datetime = Field(..., description="Timestamp of the last successful run")
    state: dict = Field(..., description="Incremental state information from the latest sync")



import logging
import os
import random
from datetime import datetime
from random import randint

import polars as pl
import pytest
from faker import Faker
from google.cloud import bigquery

from bizon.common.models import SyncMetadata
from bizon.connectors.destinations.bigquery.src.config import (
    BigQueryConfig,
    BigQueryConfigDetails,
    GCSBufferFormat,
)
from bizon.connectors.destinations.bigquery.src.destination import BigQueryDestination
from bizon.destination.config import DestinationTypes
from bizon.destination.destination import DestinationFactory
from bizon.destination.models import destination_record_schema
from bizon.monitoring.noop.monitor import NoOpMonitor

logger = logging.getLogger(__name__)

fake = Faker("en_US")

TEST_PROJECT_ID = "my_project"
TEST_TABLE_ID = "test_fake_records"
TEST_DATASET_ID = "bizon_test"
TEST_BUFFER_BUCKET = "bizon-buffer"


df_destination_records = pl.DataFrame(
    {
        "bizon_id": ["id_1", "id_2"],
        "bizon_extracted_at": [datetime(2024, 12, 5, 12, 0), datetime(2024, 12, 5, 13, 0)],
        "bizon_loaded_at": [datetime(2024, 12, 5, 12, 30), datetime(2024, 12, 5, 13, 30)],
        "source_record_id": ["record_1", "record_2"],
        "source_timestamp": [datetime(2024, 12, 5, 11, 30), datetime(2024, 12, 5, 12, 30)],
        "source_data": ["cookies", "cream"],
    },
    schema=destination_record_schema,
)


@pytest.fixture(scope="function")
def sync_metadata() -> SyncMetadata:
    return SyncMetadata(
        name="gcs_loading",
        job_id="rfou98C9DJH",
        source_name="cookie",
        stream_name="test",
        destination_name="bigquery",
        destination_alias="bigquery",
        sync_mode="full_refresh",
    )


@pytest.fixture(scope="function")
def test_table():
    table_id = f"{TEST_PROJECT_ID}.{TEST_DATASET_ID}.{TEST_TABLE_ID}"
    client = bigquery.Client()
    table = bigquery.Table(table_id)
    table = client.create_table(table)
    yield table
    client.delete_table(table_id)
    client.delete_table(f"{table_id}_temp")


@pytest.mark.skipif(
    os.getenv("POETRY_ENV_TEST") == "CI",
    reason="Skipping tests that require a BigQuery database",
)
def test_load_records_to_bigquery(my_backend_config, test_table, sync_metadata):
    bigquery_config = BigQueryConfig(
        name=DestinationTypes.BIGQUERY,
        config=BigQueryConfigDetails(
            project_id=TEST_PROJECT_ID,
            dataset_id=TEST_DATASET_ID,
            table_id=TEST_TABLE_ID,
            gcs_buffer_bucket=TEST_BUFFER_BUCKET,
            gcs_buffer_format=GCSBufferFormat.PARQUET,
        ),
    )
    fake_records = [
        {"foo": randint(0, 100), "bar": {"baz": fake.name(), "poo": float(random.randrange(155, 389)) / 100}}
        for _ in range(100)
    ]

    bq_destination = DestinationFactory().get_destination(
        sync_metadata=sync_metadata,
        config=bigquery_config,
        backend=my_backend_config,
        monitor=NoOpMonitor(sync_metadata=sync_metadata, monitoring_config=None),
    )

    assert isinstance(bq_destination, BigQueryDestination)

    success, error_msg = bq_destination.write_records(df_destination_records=df_destination_records)

    assert success is True
    assert error_msg == ""



from datetime import datetime

import polars as pl
import pytest

from bizon.connectors.destinations.bigquery.src.config import BigQueryColumn
from bizon.connectors.destinations.bigquery.src.destination import BigQueryDestination
from bizon.destination.models import destination_record_schema


def test_unnest_records_to_bigquery():
    df_destination_records = pl.DataFrame(
        data={
            "source_record_id": ["1"],
            "source_timestamp": [datetime.now()],
            "source_data": ['{"id": 1, "name": "Alice", "created_at": "2021-01-01 00:00:00"}'],
            "bizon_extracted_at": [datetime.now()],
            "bizon_loaded_at": [datetime.now()],
            "bizon_id": ["1"],
        },
        schema=destination_record_schema,
    )

    assert df_destination_records.height > 0

    res = BigQueryDestination.unnest_data(
        df_destination_records=df_destination_records,
        record_schema=[
            BigQueryColumn(name="id", type="INTEGER", mode="REQUIRED"),
            BigQueryColumn(name="name", type="STRING", mode="REQUIRED"),
            BigQueryColumn(name="created_at", type="DATETIME", mode="REQUIRED"),
        ],
    )

    assert res.height == 1


def test_unnest_records_to_bigquery_with_invalid_data():
    df_destination_records = pl.DataFrame(
        data={
            "source_record_id": ["1"],
            "source_timestamp": [datetime.now()],
            "source_data": ['{"id": 1, "name": "Alice", "created_at": "2021-01-01 00:00:00", "cookies": "chocolate"}'],
            "bizon_extracted_at": [datetime.now()],
            "bizon_loaded_at": [datetime.now()],
            "bizon_id": ["1"],
        },
        schema=destination_record_schema,
    )

    # We raise exception as the data has an extra column compared to BigQuery schema

    with pytest.raises(AssertionError) as e:
        res = BigQueryDestination.unnest_data(
            df_destination_records=df_destination_records,
            record_schema=[
                BigQueryColumn(name="id", type="INTEGER", mode="REQUIRED"),
                BigQueryColumn(name="name", type="STRING", mode="REQUIRED"),
                BigQueryColumn(name="created_at", type="DATETIME", mode="REQUIRED"),
            ],
        )



"""
Quick test script for Notion source intermediate functions.

Usage:
    # Set your token
    export NOTION_TOKEN="your_notion_integration_token"

    # Run interactively
    python -i test_notion.py

    # Then test functions:
    >>> page = source.get_page("page-id-here")
    >>> blocks = source.fetch_blocks_recursively("page-id-here")
    >>> for b in blocks[:5]:
    ...     print(source._block_to_markdown(b))
"""

import os

from bizon.connectors.sources.notion.src.config import NotionSourceConfig, NotionStreams
from bizon.connectors.sources.notion.src.source import NotionSource
from bizon.source.auth.authenticators.token import TokenAuthParams
from bizon.source.auth.config import AuthConfig, AuthType


def create_notion_source(
    token: str = None,
    page_ids: list = None,
    database_ids: list = None,
    stream: NotionStreams = NotionStreams.BLOCKS,
) -> NotionSource:
    """Create a NotionSource instance for testing."""
    token = token or os.environ.get("NOTION_TOKEN")
    if not token:
        raise ValueError("Provide token or set NOTION_TOKEN environment variable")

    config = NotionSourceConfig(
        name="notion",
        stream=stream,
        page_ids=page_ids or [],
        database_ids=database_ids or [],
        authentication=AuthConfig(
            type=AuthType.BEARER,
            params=TokenAuthParams(token=token),
        ),
        init_pipeline=False,
        max_recursion_depth=30,
    )
    return NotionSource(config)


# ==================== HELPER FUNCTIONS ====================


def get_block(source: NotionSource, block_id: str) -> dict:
    """Fetch a single block by ID."""
    response = source.session.get(f"https://api.notion.com/v1/blocks/{block_id}")
    response.raise_for_status()
    return response.json()


def get_page_markdown(source: NotionSource, page_id: str) -> str:
    """Fetch all blocks from a page and return combined markdown."""
    blocks = source.fetch_blocks_recursively(page_id, source_page_id=page_id)
    lines = []
    for block in blocks:
        md = source._block_to_markdown(block)
        if md:
            # Add indentation based on depth
            indent = "  " * block.get("depth", 0)
            lines.append(f"{indent}{md}")
    return "\n".join(lines)


def inspect_blocks(source: NotionSource, page_id: str, max_blocks: int = 10):
    """Fetch and print block details for inspection."""
    blocks = source.fetch_blocks_recursively(page_id, source_page_id=page_id)
    print(f"Found {len(blocks)} blocks")
    for i, block in enumerate(blocks[:max_blocks]):
        print(f"\n--- Block {i} ({block.get('type')}) ---")
        print(f"ID: {block.get('id')}")
        print(f"Depth: {block.get('depth')}, Order: {block.get('page_order')}")
        print(f"Markdown: {source._block_to_markdown(block)}")


def list_pages_in_database(source: NotionSource, database_id: str) -> list:
    """List all page IDs in a database."""
    return source.get_pages_from_database(database_id, apply_filter=False)


# ==================== MAIN ====================

if __name__ == "__main__":
    # Create source if token is available
    token = os.environ.get("NOTION_TOKEN")
    if token:
        source = create_notion_source(token=token)
        print("NotionSource created and available as 'source'")
        print("\nAvailable functions:")
        print("  source.get_page(page_id)")
        print("  source.get_database(database_id)")
        print("  source.get_block_children(block_id)")
        print("  source.fetch_blocks_recursively(page_id)")
        print("  source._block_to_markdown(block)")
        print("  source.search()")
        print("\nHelper functions:")
        print("  get_block(source, block_id)")
        print("  get_page_markdown(source, page_id)")
        print("  inspect_blocks(source, page_id)")
        print("  list_pages_in_database(source, database_id)")
    else:
        print("Set NOTION_TOKEN env var or call:")
        print("  source = create_notion_source(token='your_token')")



from bizon.common.models import SyncMetadata
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.monitoring.config import DatadogConfig, MonitoringConfig, MonitorType
from bizon.monitoring.datadog.monitor import DatadogMonitor
from bizon.monitoring.monitor import MonitorFactory
from bizon.monitoring.noop.monitor import NoOpMonitor

sync_metadata = SyncMetadata(
    job_id="123",
    name="pipeline_test",
    source_name="source_test",
    stream_name="stream_test",
    destination_name="destination_test",
    destination_alias="destination_test",
    sync_mode="full_refresh",
)


def test_datadog_monitor():
    dd_monitor = MonitorFactory.get_monitor(
        sync_metadata=sync_metadata,
        monitoring_config=MonitoringConfig(
            type=MonitorType.DATADOG, config=DatadogConfig(datadog_agent_host="localhost", datadog_agent_port=8125)
        ),
    )

    assert type(dd_monitor) == DatadogMonitor
    assert dd_monitor.monitoring_config.type == MonitorType.DATADOG


def test_datadog_track_records_synced():
    dd_monitor = MonitorFactory.get_monitor(
        sync_metadata=sync_metadata,
        monitoring_config=MonitoringConfig(
            type=MonitorType.DATADOG, config=DatadogConfig(datadog_agent_host="localhost", datadog_agent_port=8125)
        ),
    )
    dd_monitor.track_records_synced(num_records=10, destination_id="123")


def test_datadog_track_pipeline_status():
    dd_monitor = MonitorFactory.get_monitor(
        sync_metadata=sync_metadata,
        monitoring_config=MonitoringConfig(
            type=MonitorType.DATADOG, config=DatadogConfig(datadog_agent_host="localhost", datadog_agent_port=8125)
        ),
    )
    dd_monitor.track_pipeline_status(PipelineReturnStatus.SUCCESS)


def test_datadog_track_large_records_synced():
    dd_monitor = MonitorFactory.get_monitor(
        sync_metadata=sync_metadata,
        monitoring_config=MonitoringConfig(
            type=MonitorType.DATADOG, config=DatadogConfig(datadog_agent_host="localhost", datadog_agent_port=8125)
        ),
    )
    dd_monitor.track_large_records_synced(num_records=10)


def test_no_op_monitor():
    no_op_monitor = MonitorFactory.get_monitor(sync_metadata=sync_metadata, monitoring_config=None)

    assert type(no_op_monitor) == NoOpMonitor
    assert no_op_monitor.monitoring_config is None



import os
import time
from contextlib import contextmanager

import pytest
from confluent_kafka import KafkaException

from bizon.connectors.sources.kafka.src.config import (
    KafkaAuthConfig,
    KafkaSourceConfig,
    MessageEncoding,
    TopicConfig,
)
from bizon.connectors.sources.kafka.src.source import KafkaSource
from bizon.source.auth.authenticators.basic import BasicHttpAuthParams
from bizon.source.auth.config import AuthType

# Skip all tests in this module unless KAFKA_E2E_TESTS environment variable is set
pytestmark = pytest.mark.skipif(
    not os.getenv("KAFKA_E2E_TESTS"), reason="Kafka e2e tests require KAFKA_E2E_TESTS environment variable to be set"
)


@contextmanager
def timeout_context(seconds):
    """Context manager to track execution time"""
    start_time = time.time()
    yield
    elapsed = time.time() - start_time
    assert elapsed < seconds + 5, f"Operation took {elapsed:.2f}s, expected < {seconds + 5}s"


class TestKafkaAuthenticationFailures:
    """Test authentication failure scenarios for Kafka source"""

    @pytest.fixture
    def kafka_host(self):
        """Get Kafka host from environment or use localhost"""
        return os.getenv("KAFKA_HOST", "localhost:9092")

    @pytest.fixture
    def invalid_auth_config(self, kafka_host):
        """Config with invalid authentication credentials"""
        return KafkaSourceConfig(
            name="kafka",
            stream="topic",
            topics=[TopicConfig(name="test-topic", destination_id="test_dest")],
            bootstrap_servers=kafka_host,
            group_id="test-auth-failures",
            consumer_timeout=5,  # Short timeout for faster test
            message_encoding=MessageEncoding.UTF_8,
            authentication=KafkaAuthConfig(
                type=AuthType.BASIC,
                params=BasicHttpAuthParams(username="invalid_user", password="invalid_password"),
            ),
        )

    @pytest.fixture
    def unreachable_server_config(self):
        """Config pointing to unreachable Kafka server"""
        return KafkaSourceConfig(
            name="kafka",
            stream="topic",
            topics=[TopicConfig(name="test-topic", destination_id="test_dest")],
            bootstrap_servers="unreachable-kafka:9092",
            group_id="test-unreachable",
            consumer_timeout=3,  # Very short timeout
            message_encoding=MessageEncoding.UTF_8,
            authentication=KafkaAuthConfig(
                type=AuthType.BASIC,
                params=BasicHttpAuthParams(username="test_user", password="test_password"),
            ),
        )

    @pytest.fixture
    def wrong_port_config(self, kafka_host):
        """Config with wrong port"""
        host = kafka_host.split(":")[0]  # Extract host without port
        return KafkaSourceConfig(
            name="kafka",
            stream="topic",
            topics=[TopicConfig(name="test-topic", destination_id="test_dest")],
            bootstrap_servers=f"{host}:9999",  # Wrong port
            group_id="test-wrong-port",
            consumer_timeout=3,
            message_encoding=MessageEncoding.UTF_8,
            authentication=KafkaAuthConfig(
                type=AuthType.BASIC,
                params=BasicHttpAuthParams(username="test_user", password="test_password"),
            ),
        )

    def test_invalid_authentication_credentials(self, invalid_auth_config):
        """Test that invalid authentication credentials are properly handled with timeout"""
        source = KafkaSource(invalid_auth_config)

        with timeout_context(10):  # Should complete within 10 seconds due to timeout
            success, error_msg = source.check_connection()

        assert not success, "Connection should fail with invalid credentials"
        assert error_msg is not None, "Error message should be provided"
        assert "Kafka connection failed" in error_msg

    def test_unreachable_kafka_server(self, unreachable_server_config):
        """Test connection to unreachable Kafka server times out properly"""
        source = KafkaSource(unreachable_server_config)

        with timeout_context(8):  # Should timeout within 8 seconds (3s timeout + buffer)
            success, error_msg = source.check_connection()

        assert not success, "Connection should fail for unreachable server"
        assert error_msg is not None, "Error message should be provided"
        assert "Kafka connection failed" in error_msg

    def test_wrong_port_connection(self, wrong_port_config):
        """Test connection to wrong port times out properly"""
        source = KafkaSource(wrong_port_config)

        with timeout_context(8):  # Should timeout within 8 seconds
            success, error_msg = source.check_connection()

        assert not success, "Connection should fail for wrong port"
        assert error_msg is not None, "Error message should be provided"
        assert "Kafka connection failed" in error_msg

    def test_get_data_with_invalid_auth(self, invalid_auth_config):
        """Test that get() method properly handles authentication failures and pipeline stops"""
        source = KafkaSource(invalid_auth_config)

        with timeout_context(10):
            # Ensure that KafkaException is raised, which should stop the pipeline
            with pytest.raises(KafkaException) as exc_info:
                source.get()

            # Verify the exception contains authentication/connection failure information
            exception_str = str(exc_info.value)
            assert any(
                keyword in exception_str.lower()
                for keyword in ["authentication", "sasl", "ssl", "transport", "broker", "connection"]
            ), f"Exception should indicate auth/connection failure: {exception_str}"

            # Verify the exception is fatal (should stop pipeline)
            kafka_error = exc_info.value.args[0] if exc_info.value.args else None
            if kafka_error and hasattr(kafka_error, "fatal"):
                # If it's a fatal error, pipeline should definitely stop
                if kafka_error.fatal():
                    assert True, "Fatal error correctly stops pipeline"

            # At minimum, any KafkaException during get() should stop the pipeline
            assert True, "KafkaException raised - pipeline will stop as expected"

    def test_get_data_with_unreachable_server(self, unreachable_server_config):
        """Test that get() method properly handles unreachable server and pipeline stops"""
        source = KafkaSource(unreachable_server_config)

        with timeout_context(8):
            # Ensure that KafkaException is raised, which should stop the pipeline
            with pytest.raises(KafkaException) as exc_info:
                source.get()

            # Verify the exception contains connection failure information
            exception_str = str(exc_info.value)
            assert any(
                keyword in exception_str.lower()
                for keyword in ["transport", "broker", "connection", "timeout", "network", "unreachable"]
            ), f"Exception should indicate connection failure: {exception_str}"

            # Any KafkaException during get() should stop the pipeline
            assert True, "KafkaException raised - pipeline will stop as expected"

    def test_timeout_configuration_respected(self, unreachable_server_config):
        """Test that consumer timeout configuration is respected"""
        # Set very short timeout
        unreachable_server_config.consumer_timeout = 1
        source = KafkaSource(unreachable_server_config)

        start_time = time.time()
        success, _ = source.check_connection()
        elapsed = time.time() - start_time

        assert not success
        # Should complete within reasonable time (timeout + processing overhead)
        assert elapsed < 5, f"Operation took {elapsed:.2f}s, should respect 1s timeout"

    def test_nonexistent_topic_handling(self, kafka_host):
        """Test handling of non-existent topics"""
        config = KafkaSourceConfig(
            name="kafka",
            stream="topic",
            topics=[TopicConfig(name="non-existent-topic-12345", destination_id="test_dest")],
            bootstrap_servers=kafka_host,
            group_id="test-nonexistent-topic",
            consumer_timeout=5,
            message_encoding=MessageEncoding.UTF_8,
            consumer_config={
                "auto.offset.reset": "earliest",
                "enable.auto.commit": False,
                "security.protocol": "PLAINTEXT",
            },
            authentication=KafkaAuthConfig(
                type=AuthType.BASIC,
                params=BasicHttpAuthParams(username="test_user", password="test_password"),
            ),
        )

        source = KafkaSource(config)

        with timeout_context(10):
            success, error_msg = source.check_connection()

        # Depending on Kafka setup, this might succeed (auto-create) or fail
        # We mainly want to ensure it doesn't hang and returns within timeout
        if not success:
            assert error_msg is not None
            # Could be "not found" or connection error depending on setup
            assert "not found" in error_msg or "transport failure" in error_msg or "connection" in error_msg.lower()

    @pytest.mark.parametrize("timeout_value", [1, 3, 5])
    def test_different_timeout_values(self, unreachable_server_config, timeout_value):
        """Test that different timeout values are respected"""
        unreachable_server_config.consumer_timeout = timeout_value
        source = KafkaSource(unreachable_server_config)

        start_time = time.time()
        success, _ = source.check_connection()
        elapsed = time.time() - start_time

        assert not success
        # Allow some overhead but ensure timeout is roughly respected
        assert elapsed < timeout_value + 3, f"Operation took {elapsed:.2f}s, timeout was {timeout_value}s"

    def test_malformed_bootstrap_servers(self):
        """Test handling of malformed bootstrap servers"""
        config = KafkaSourceConfig(
            name="kafka",
            stream="topic",
            topics=[TopicConfig(name="test-topic", destination_id="test_dest")],
            bootstrap_servers="invalid-format-server",  # Missing port
            group_id="test-malformed",
            consumer_timeout=3,
            message_encoding=MessageEncoding.UTF_8,
            authentication=KafkaAuthConfig(
                type=AuthType.BASIC,
                params=BasicHttpAuthParams(username="test_user", password="test_password"),
            ),
        )

        source = KafkaSource(config)

        with timeout_context(8):
            success, error_msg = source.check_connection()

        assert not success, "Connection should fail with malformed server address"
        assert error_msg is not None

    def test_pipeline_integration_auth_failure_stops_execution(self, invalid_auth_config):
        """Integration test: Verify authentication failure stops pipeline execution immediately"""
        source = KafkaSource(invalid_auth_config)

        # Test that check_connection fails first (pre-flight check)
        with timeout_context(10):
            connection_success, connection_error = source.check_connection()

        assert not connection_success, "Connection check should fail with invalid auth"
        assert connection_error is not None, "Connection error should be reported"

        # Test that get() method also fails and raises exception (runtime check)
        with timeout_context(10):
            with pytest.raises(KafkaException) as exc_info:
                # This simulates what happens when pipeline tries to fetch data
                source.get()

            # Ensure the exception would stop pipeline execution
            exception_str = str(exc_info.value)
            assert len(exception_str) > 0, "Exception message should not be empty"

            # Log that pipeline would stop
            print(f"✅ Pipeline correctly stopped with exception: {exception_str[:100]}...")

        # Verify that no partial data is returned when auth fails
        # This ensures data integrity - either all data or no data, never partial
        assert True, "Authentication failure correctly prevents partial data extraction"



import os

from bizon.cli.utils import parse_from_yaml
from bizon.engine.engine import RunnerFactory

config = parse_from_yaml(os.path.abspath("bizon/connectors/sources/periscope/config/periscope_dashboards.yml"))

runner = RunnerFactory.create_from_config_dict(config=config)
runner.run()



import concurrent.futures
import time
import traceback

from loguru import logger

from bizon.engine.runner.runner import AbstractRunner


class ProcessRunner(AbstractRunner):
    def __init__(self, config: dict):
        super().__init__(config)

    # TODO: refacto this
    def get_kwargs(self):
        if self.bizon_config.engine.queue.type == "python_queue":
            from multiprocessing import Manager

            manager = Manager()
            queue = manager.Queue(maxsize=self.bizon_config.engine.queue.config.queue.max_size)
            return {"queue": queue}

        return {}

    def run(self):
        """Run the pipeline with dedicated threads for source and destination"""

        extra_kwargs = self.get_kwargs()
        job = AbstractRunner.init_job(bizon_config=self.bizon_config, config=self.config, **extra_kwargs)

        # Store the future results
        result_producer = None
        result_consumer = None

        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.bizon_config.engine.runner.config.max_workers
        ) as executor:
            future_producer = executor.submit(
                AbstractRunner.instanciate_and_run_producer,
                self.bizon_config,
                self.config,
                job.id,
                **extra_kwargs,
            )
            logger.info("Producer process has started ...")

            time.sleep(self.bizon_config.engine.runner.config.consumer_start_delay)

            future_consumer = executor.submit(
                AbstractRunner.instanciate_and_run_consumer,
                self.bizon_config,
                self.config,
                job.id,
                **extra_kwargs,
            )
            logger.info("Consumer process has started ...")

            self._is_running = True

            while future_producer.running() and future_consumer.running():
                logger.debug("Producer and consumer are still running ...")
                self._is_running = True
                time.sleep(self.bizon_config.engine.runner.config.is_alive_check_interval)

            self._is_running = False

            if not future_producer.running():
                result_producer = future_producer.result()
                logger.info(f"Producer process stopped running with result: {result_producer}")

                if result_producer.SUCCESS:
                    logger.info("Producer thread has finished successfully, will wait for consumer to finish ...")
                else:
                    logger.error("Producer thread failed, stopping consumer ...")
                    executor.shutdown(wait=False)

            if not future_consumer.running():
                try:
                    future_consumer.result()
                except Exception as e:
                    logger.error(f"Consumer thread stopped running with error {e}")
                    logger.error(traceback.format_exc())
                finally:
                    executor.shutdown(wait=False)

        return True



from typing import Union

from pydantic import BaseModel, ConfigDict, Field

from .backend.adapters.sqlalchemy.config import (
    BigQuerySQLAlchemyConfig,
    PostgresSQLAlchemyConfig,
    SQLiteConfigDetails,
    SQLiteInMemoryConfig,
    SQLiteSQLAlchemyConfig,
)
from .backend.config import BackendTypes
from .queue.adapters.kafka.config import KafkaConfig
from .queue.adapters.python_queue.config import (
    PythonQueueConfig,
    PythonQueueConfigDetails,
    PythonQueueConsumerConfig,
    PythonQueueQueueConfig,
)
from .queue.adapters.rabbitmq.config import RabbitMQConfig
from .queue.config import QueueTypes
from .runner.config import RunnerConfig, RunnerFuturesConfig, RunnerTypes


class EngineConfig(BaseModel):
    # Forbid extra keys in the model
    model_config = ConfigDict(extra="forbid")

    backend: Union[
        PostgresSQLAlchemyConfig,
        SQLiteInMemoryConfig,
        SQLiteSQLAlchemyConfig,
        BigQuerySQLAlchemyConfig,
    ] = Field(
        description="Configuration for the backend",
        default=SQLiteSQLAlchemyConfig(
            type=BackendTypes.SQLITE,
            config=SQLiteConfigDetails(
                database="bizon",
                schema="NOT_USED_IN_SQLITE",
            ),
            syncCursorInDBEvery=2,
        ),
        discriminator="type",
    )

    queue: Union[
        KafkaConfig,
        RabbitMQConfig,
        PythonQueueConfig,
    ] = Field(
        description="Configuration for the queue",
        default=PythonQueueConfig(
            type=QueueTypes.PYTHON_QUEUE,
            config=PythonQueueConfigDetails(
                queue=PythonQueueQueueConfig(max_size=0),
                consumer=PythonQueueConsumerConfig(poll_interval=2),
            ),
        ),
        discriminator="type",
    )

    runner: RunnerConfig = Field(
        description="Runner to use for the pipeline",
        default=RunnerConfig(
            type=RunnerTypes.THREAD,
            config=RunnerFuturesConfig(),
            log_level="INFO",
        ),
    )



from typing import Literal

from pydantic import BaseModel, Field

from bizon.engine.queue.config import (
    AbastractQueueConfigDetails,
    AbstractQueueConfig,
    QueueTypes,
)


class PythonQueueQueueConfig(BaseModel):
    max_size: int = Field(0, description="Maximum size of the queue, by default 0 means unlimited")


class PythonQueueConsumerConfig(BaseModel):
    poll_interval: int = Field(1, description="Interval in seconds to poll the queue in seconds")


class PythonQueueConfigDetails(AbastractQueueConfigDetails):
    queue: PythonQueueQueueConfig = Field(
        default=PythonQueueQueueConfig(max_size=0), description="Configuration of the queue"
    )
    consumer: PythonQueueConsumerConfig = Field(
        default=PythonQueueConsumerConfig(poll_interval=1), description="Kafka consumer configuration"
    )


class PythonQueueConfig(AbstractQueueConfig):
    type: Literal[QueueTypes.PYTHON_QUEUE]
    config: PythonQueueConfigDetails



from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from bizon.alerting.models import AlertingConfig
from bizon.connectors.destinations.bigquery.src.config import (
    BigQueryColumn,
    BigQueryConfig,
)
from bizon.connectors.destinations.bigquery_streaming.src.config import (
    BigQueryStreamingConfig,
)
from bizon.connectors.destinations.bigquery_streaming_v2.src.config import (
    BigQueryStreamingV2Config,
)
from bizon.connectors.destinations.file.src.config import FileDestinationConfig
from bizon.connectors.destinations.logger.src.config import LoggerConfig
from bizon.engine.config import EngineConfig
from bizon.monitoring.config import MonitoringConfig
from bizon.source.config import SourceConfig, SourceSyncModes
from bizon.transform.config import TransformModel


class StreamSourceConfig(BaseModel):
    """Source-specific stream routing configuration.

    Uses extra='allow' to support source-specific fields like:
    - topic (Kafka)
    - endpoint (API sources)
    - channel (other streaming sources)
    """

    model_config = ConfigDict(extra="allow")

    # Common field for stream identifier
    name: Optional[str] = Field(None, description="Stream identifier within the source")

    # Kafka-specific
    topic: Optional[str] = Field(None, description="Kafka topic name")

    # API-specific
    endpoint: Optional[str] = Field(None, description="API endpoint path")


class StreamDestinationConfig(BaseModel):
    """Destination configuration for a stream.

    Supports destination-specific schema definitions and options.
    Uses extra='allow' to support destination-specific overrides.
    """

    model_config = ConfigDict(extra="allow")

    # Universal destination identifier
    table_id: str = Field(..., description="Full destination identifier (e.g., project.dataset.table)")

    # BigQuery-specific schema (can be extended for other destinations)
    record_schema: Optional[list[BigQueryColumn]] = Field(None, description="Schema for the destination records")
    clustering_keys: Optional[list[str]] = Field(None, description="Clustering keys for the destination table")


class StreamConfig(BaseModel):
    """Configuration for a single stream.

    Consolidates source stream routing and destination configuration in one place,
    eliminating duplication of destination_id between source and destination configs.
    """

    model_config = ConfigDict(extra="forbid")

    name: str = Field(..., description="Logical name for this stream")
    source: StreamSourceConfig = Field(..., description="Source-specific routing configuration")
    destination: StreamDestinationConfig = Field(
        ..., description="Destination configuration including table and schema"
    )

    @field_validator("destination")
    @classmethod
    def validate_table_id_format(cls, v: StreamDestinationConfig) -> StreamDestinationConfig:
        """Ensure table_id follows expected format for BigQuery-like destinations."""
        if v.table_id:
            parts = v.table_id.split(".")
            if len(parts) != 3:
                raise ValueError(
                    f"table_id must be in format 'project.dataset.table', got: {v.table_id}. "
                    f"Found {len(parts)} parts instead of 3."
                )
        return v


class BizonConfig(BaseModel):
    # Forbid extra keys in the model
    model_config = ConfigDict(extra="forbid")

    # Unique name to identify the sync configuration
    name: str = Field(..., description="Unique name for this sync configuration")

    source: SourceConfig = Field(
        description="Source configuration",
        default=...,
    )

    transforms: Optional[list[TransformModel]] = Field(
        description="List of transformations to apply to the source data",
        default=[],
    )

    destination: Union[
        BigQueryConfig,
        BigQueryStreamingConfig,
        BigQueryStreamingV2Config,
        LoggerConfig,
        FileDestinationConfig,
    ] = Field(
        description="Destination configuration",
        discriminator="name",
        default=...,
    )

    engine: EngineConfig = Field(
        description="Engine configuration",
        default=EngineConfig(),
    )

    alerting: Optional[AlertingConfig] = Field(
        description="Alerting configuration",
        default=None,
    )

    monitoring: Optional[MonitoringConfig] = Field(
        description="Monitoring configuration",
        default=None,
    )

    streams: Optional[list[StreamConfig]] = Field(
        None,
        description="Stream routing configuration (opt-in for multi-table streaming). "
        "Consolidates source stream definitions with destination tables and schemas.",
    )

    @field_validator("streams")
    @classmethod
    def validate_streams_config(cls, v: Optional[list[StreamConfig]], info) -> Optional[list[StreamConfig]]:
        """Validate streams configuration consistency."""
        if not v:
            return v

        # Check for duplicate stream names
        names = [s.name for s in v]
        if len(names) != len(set(names)):
            duplicates = [name for name in names if names.count(name) > 1]
            raise ValueError(f"Duplicate stream names found in streams configuration: {set(duplicates)}")

        # Check for duplicate table_ids
        table_ids = [s.destination.table_id for s in v]
        if len(table_ids) != len(set(table_ids)):
            duplicates = [tid for tid in table_ids if table_ids.count(tid) > 1]
            raise ValueError(f"Duplicate table_ids found in streams configuration: {set(duplicates)}")

        # Validate that source sync_mode is 'stream' if streams config is used
        source_config = info.data.get("source") if info.data else None
        if source_config and source_config.sync_mode != SourceSyncModes.STREAM:
            raise ValueError(
                f"Configuration Error: 'streams' configuration requires source.sync_mode='stream'. "
                f"Current sync_mode: {source_config.sync_mode}. "
                f"Please update your config to use:\n"
                f"  source:\n"
                f"    sync_mode: stream"
            )

        return v

    @model_validator(mode="before")
    @classmethod
    def inject_config_from_streams(cls, data: Any) -> Any:
        """Inject source and destination config from streams.

        This runs BEFORE field validation, enriching both source and destination
        configs from the streams configuration. This allows:
        1. Sources (like Kafka) to omit topics - they're extracted from streams
        2. Destinations with unnest=true to work without duplicate record_schemas

        This is source-agnostic: each source type can extract what it needs from streams.
        """
        if not isinstance(data, dict):
            return data

        streams = data.get("streams")
        if not streams:
            return data

        source = data.get("source")
        if source and isinstance(source, dict):
            source_name = source.get("name")

            # Kafka: inject topics from streams
            if source_name == "kafka":
                # Check if topics is missing, None, or empty list
                if not source.get("topics") or source.get("topics") == []:
                    topics = []
                    for stream in streams:
                        if isinstance(stream, dict):
                            stream_src = stream.get("source", {})
                            stream_dest = stream.get("destination", {})
                            if stream_src.get("topic"):
                                topics.append(
                                    {
                                        "name": stream_src.get("topic"),
                                        "destination_id": stream_dest.get("table_id", ""),
                                    }
                                )
                    if topics:
                        source["topics"] = topics

        destination = data.get("destination")
        if not destination or not isinstance(destination, dict):
            return data

        destination_config = destination.get("config")
        if not destination_config or not isinstance(destination_config, dict):
            return data

        # Only inject if record_schemas is not already set or is empty
        if not destination_config.get("record_schemas"):
            # Build record_schemas from streams
            record_schemas = []
            for stream in streams:
                if isinstance(stream, dict):
                    stream_dest = stream.get("destination", {})
                    if stream_dest.get("record_schema"):
                        record_schema_config = {
                            "destination_id": stream_dest.get("table_id"),
                            "record_schema": stream_dest.get("record_schema"),
                            "clustering_keys": stream_dest.get("clustering_keys"),
                        }
                        record_schemas.append(record_schema_config)

            # Inject into destination config
            if record_schemas:
                destination_config["record_schemas"] = record_schemas

        return data


class SyncMetadata(BaseModel):
    """Model which stores general metadata around a sync.
    Facilitate usage of basic info across entities
    """

    name: str
    job_id: str
    source_name: str
    stream_name: str
    sync_mode: SourceSyncModes
    destination_name: str
    destination_alias: str

    @classmethod
    def from_bizon_config(cls, job_id: str, config: BizonConfig) -> "SyncMetadata":
        return cls(
            name=config.name,
            job_id=job_id,
            source_name=config.source.name,
            stream_name=config.source.stream,
            sync_mode=config.source.sync_mode,
            destination_name=config.destination.name,
            destination_alias=config.destination.alias,
        )



import os

from yaml import safe_load

from bizon.cli.utils import parse_from_yaml
from bizon.common.models import SyncMetadata
from bizon.connectors.destinations.logger.src.destination import LoggerDestination
from bizon.engine.backend.config import BackendTypes
from bizon.engine.engine import RunnerFactory
from bizon.engine.queue.adapters.kafka.queue import KafkaQueue
from bizon.engine.queue.adapters.rabbitmq.queue import RabbitMQ
from bizon.monitoring.noop.monitor import NoOpMonitor
from bizon.source.callback import NoOpSourceCallback


def test_parse_from_yaml():
    config = parse_from_yaml(os.path.abspath("tests/engine/dummy_pipeline_sqlite.yml"))
    runner = RunnerFactory.create_from_config_dict(config=config)
    assert runner.is_running is False


def test_parse_task_runner_python_queue():
    config_yaml = """
        name: test_job

        source:
            name: dummy
            stream: creatures
            authentication:
                type: api_key
                params:
                    token: dummy_key

        destination:
            name: logger
            config:
                dummy: dummy

        engine:
            backend:
                type: sqlite_in_memory
                config:
                    database: not_used
                    schema: not_used
                    syncCursorInDBEvery: 400
            runner:
                log_level: INFO
        """
    config = safe_load(config_yaml)

    runner = RunnerFactory.create_from_config_dict(config=config)

    backend = runner.get_backend(bizon_config=runner.bizon_config)

    destination = runner.get_destination(
        bizon_config=runner.bizon_config,
        backend=backend,
        job_id="123",
        source_callback=NoOpSourceCallback(config={}),
        monitor=NoOpMonitor(
            sync_metadata=SyncMetadata.from_bizon_config(job_id="123", config=runner.bizon_config),
            monitoring_config=None,
        ),
    )

    assert isinstance(destination, LoggerDestination)

    # Check if the backend per default is of type sqlite
    assert backend.type == BackendTypes.SQLITE_IN_MEMORY

    assert runner.bizon_config.engine.runner.log_level == "INFO"

    assert backend.config.syncCursorInDBEvery == 400


def test_parse_task_runner_kafka_queue():
    config_yaml = """
        name: test_job

        source:
            name: dummy
            stream: creatures
            authentication:
                type: api_key
                params:
                    token: dummy_key

        destination:
            name: logger
            config:
                dummy: dummy

        engine:
            backend:
                type: sqlite_in_memory
                config:
                    database: not_used
                    schema: not_used
                    syncCursorInDBEvery: 200

            queue:
                type: kafka
                config:
                    queue:
                        bootstrap_server: localhost:9092

            runner:
                log_level: DEBUG
        """
    config = safe_load(config_yaml)
    runner = RunnerFactory.create_from_config_dict(config=config)

    backend = runner.get_backend(bizon_config=runner.bizon_config)

    destination = runner.get_destination(
        bizon_config=runner.bizon_config,
        backend=backend,
        job_id="123",
        source_callback=NoOpSourceCallback(config={}),
        monitor=NoOpMonitor(
            sync_metadata=SyncMetadata.from_bizon_config(job_id="123", config=runner.bizon_config),
            monitoring_config=None,
        ),
    )

    queue = runner.get_queue(bizon_config=runner.bizon_config)

    assert isinstance(destination, LoggerDestination)

    assert backend.type == BackendTypes.SQLITE_IN_MEMORY

    assert isinstance(queue, KafkaQueue)
    assert queue.config.queue.bootstrap_server == "localhost:9092"


def test_parse_task_runner_rabbitmq_queue():
    config_yaml = """

        name: test_job

        source:
            name: dummy
            stream: creatures
            authentication:
                type: api_key
                params:
                    token: dummy_key

        destination:
            name: logger
            config:
                dummy: dummy

        engine:
            backend:
                type: sqlite_in_memory
                config:
                    database: not_used
                    schema: not_used
                    syncCursorInDBEvery: 200

            queue:
                type: rabbitmq
                config:
                    queue:
                        host: localhost
                        port: 5672
                        queue_name: bizon
                    consumer:
                        poll_interval: 1
            runner:
                log_level: DEBUG
        """
    config = safe_load(config_yaml)
    runner = RunnerFactory.create_from_config_dict(config=config)

    assert runner.bizon_config.name == "test_job"

    backend = runner.get_backend(bizon_config=runner.bizon_config)

    destination = runner.get_destination(
        bizon_config=runner.bizon_config,
        backend=backend,
        job_id="123",
        source_callback=NoOpSourceCallback(config={}),
        monitor=NoOpMonitor(
            sync_metadata=SyncMetadata.from_bizon_config(job_id="123", config=runner.bizon_config),
            monitoring_config=None,
        ),
    )

    queue = runner.get_queue(bizon_config=runner.bizon_config)

    assert isinstance(destination, LoggerDestination)

    assert backend.type == BackendTypes.SQLITE_IN_MEMORY

    assert isinstance(queue, RabbitMQ)
    assert queue.config.queue.queue_name == "bizon"
    assert queue.config.queue.host == "localhost"
    assert queue.config.queue.port == 5672
    assert queue.config.consumer.poll_interval == 1



import tempfile

from click.testing import CliRunner

from bizon.cli.main import cli

BIZON_CONFIG_DUMMY_TO_FILE = """
name: test_job_3

source:
    name: dummy
    stream: creatures

    authentication:
        type: api_key
        params:
            token: dummy_key
    sleep: 2

destination:
    name: file
    config:
        destination_id: test.jsonl
        format: json

transforms:
- label: transform_data
  python: |
    if 'name' in data:
        data['name'] = data['this_key_doesnt_exist'].upper()
"""


def test_e2e_run_command_dummy_to_file():
    runner = CliRunner()

    with tempfile.NamedTemporaryFile(delete=False) as temp:
        # Write config in temp file
        with open(temp.name, "w") as f:
            f.write(BIZON_CONFIG_DUMMY_TO_FILE)

        runner = CliRunner()
        result = runner.invoke(cli, ["run", temp.name])

        assert result.exit_code == 1
        assert (
            "Pipeline finished with status Failure (Producer: killed_by_runner, Consumer: transform_error)"
            in result.stderr
        )
